# -*- coding: utf-8 -*-

from .serializable import Serializable
from .eq import DeepEqual
from .classes import Classes

from collections import OrderedDict
import builtins

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))

""" Allowed data (Parameter and Dataset) types and the corresponding classe names.
The keys are mnemonics for humans; the values are type(x).__name__.
"""
DataTypes = {
    'binary': 'int',
    'integer': 'int',
    'short': 'int',
    'byte': 'int',
    'hex': 'int',
    'float': 'float',
    'string': 'str',
    'boolean': 'bool',
    'finetime': 'FineTime',
    'finetime1': 'FineTime1',
    'baseProduct': 'BaseProduct',
    'mapContext': 'MapContext',
    'product': 'Product',
    'vector': 'Vector',
    'vector2d': 'Vector2D',
    'quaternion': 'Quaternion',
    '': 'None'
}


""" maps class type names to human-friendly types """
DataTypeNames = {}
for tn, tt in DataTypes.items():
    if tt == 'int':
        DataTypeNames[tt] = 'integer'
    else:
        DataTypeNames[tt] = tn
DataTypeNames.update({
    'NoneType': '',
    'dict': 'vector',
    'ODict': 'vector'
})
del tt, tn


def cast(val, typ_, namespace=None):
    """ casts the input value to type specified, which is in DataTypeNames.

    For example 'binary' type '0x9' is casted to int 9.

    namespace: default is Classes.mapping.
    """
    t = DataTypes[typ_]
    bd = builtins.__dict__
    vstring = str(val).lower()
    if t in bd.keys():
        if t == 'int':
            base = 16 if vstring.startswith(
                '0x') else 2 if vstring.startswith('0b') else 10
            return bd[t](vstring, base)
        return bd[t](val)
    else:
        return Classes.mapping[t](val) if namespace is None else namespace[t](val)


class Vector(Serializable, DeepEqual):
    """ N dimensional vector.

    If unit, description, type etc meta data is needed, use a Parameter.
    """

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0, 0] components.
        """
        if components is None:
            self._data = [0, 0, 0]
        else:
            self.setComponents(components)
        super(Vector, self).__init__(**kwds)

    @property
    def components(self):
        """ for property getter
        """
        return self.getComponents()

    @components.setter
    def components(self, components):
        """ for property setter
        """
        self.setComponents(components)

    def getComponents(self):
        """ Returns the actual components that is allowed for the components
        of this vector."""
        return self._data

    def setComponents(self, components):
        """ Replaces the current components of this vector. """
        # for c in components:
        #     if not isinstance(c, Number):
        #         raise TypeError('Components must all be numbers.')
        # must be list to make round-trip Json
        self._data = list(components)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return str(self._data)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(
            components=list(self._data),
            _STID=self._STID)


class Vector2D(Vector):
    """ Vector with 2-component data.
    """

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0] components

        """

        super(Vector2D, self).__init__(**kwds)

        if components is None:
            self._data = [0, 0]
        else:
            self.setComponents(components)


class Quaternion(Vector):
    """ Quaternion with 4-component data.
    """

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0, 0, 0] components

        """

        super(Quaternion, self).__init__(**kwds)

        if components is None:
            self._data = [0, 0, 0, 0]
        else:
            self.setComponents(components)



# -*- coding: utf-8 -*-

from ..utils.common import trbk
from ..utils.moduleloader import SelectiveMetaFinder, installSelectiveMetaFinder

import sys
import logging
import copy
import importlib

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


''' Note: this has to be in a different file where other interface
classes are defined to avoid circular dependency (such as ,
Serializable.
'''


class Classes_meta(type):
    """ metaclass for 'classproperty'.
        # https://stackoverflow.com/a/1800999
    """
    # modules and classes to import from them
    module_class = {
        'fdi.dataset.deserialize': ['deserialize'],
        'fdi.dataset.listener': ['ListnerSet'],
        'fdi.dataset.serializable': ['Serializable'],
        'fdi.dataset.eq': ['DeepEqual'],
        'fdi.dataset.odict': ['ODict'],
        'fdi.dataset.finetime': ['FineTime', 'FineTime1', 'utcobj'],
        'fdi.dataset.history': ['History'],
        'fdi.dataset.baseproduct': ['BaseProduct'],
        'fdi.dataset.product': ['Product'],
        'fdi.dataset.testproducts': ['TP', 'TC', 'TM'],
        'fdi.dataset.datatypes': ['Vector', 'Vector2D', 'Quaternion'],
        'fdi.dataset.metadata': ['AbstractParameter', 'Parameter', 'NumericParameter', 'DateParameter', 'StringParameter', 'MetaData'],
        'fdi.dataset.dataset': ['GenericDataset', 'ArrayDataset',
                                'TableDataset', 'CompositeDataset', 'Column'],
        'fdi.dataset.product.readonlydict': ['ReadOnlyDict'],
        'fdi.pal.context': ['AbstractContext', 'Context',
                            'AbstractMapContext', 'MapContext',
                            'RefContainer',
                            'ContextRuleException'],
        'fdi.pal.urn': ['Urn'],
        'fdi.pal.productref': ['ProductRef']
    }

    # class list from the package
    _package = {}
    # class list with modifcation
    _classes = {}

    def __init__(cls, *args, **kwds):
        """ Class is initialized with built-in classes by default.
        """
        super().__init__(*args, **kwds)

    def updateMapping(cls, c=None, rerun=False, exclude=None, ignore_missing=False, verbose=False, ignore_error=False):
        """ Updates classes mapping.
        Make the package mapping if it has not been made.
        """
        if exclude is None:
            exclude = []
        try:
            cls.importModuleClasses(
                rerun=rerun, exclude=exclude, verbose=verbose)
        except (ModuleNotFoundError, SyntaxError) as e:
            if ignore_error:
                logger.warning('!'*80 +
                               '\nUnable to import "%s" module. Ignored\n' % clp +
                               '!'*80+'\n'+str(e)+'\n'+'!'*80)
            else:
                raise

        # cls._classes.clear()
        cls._classes.update(copy.copy(cls._package))
        if c:
            cls._classes.update(c)
        return cls._classes

    def importModuleClasses(cls, rerun=False, exclude=None, verbose=False):
        """ The set of deserializable classes in module_class is maintained by hand.

        Do nothing if the classes mapping is already made so repeated calls will not cost  more time.

        rerun: set to True to force re-import. If the module-class list has never been imported, it will be imported regardless rerun.
        exclude: modules whose names (without '.') are in exclude are not imported.
        """

        if len(cls._package) and not rerun:
            return
        if exclude is None:
            exclude = []

        cls._package.clear()
        SelectiveMetaFinder.exclude = exclude
        msg = 'With %s excluded.. and SelectiveMetaFinder.exclude=%s' % (
            str(exclude), str(SelectiveMetaFinder.exclude))
        if verbose:
            logger.info(msg)
        else:
            logger.debug(msg)

        for module_name, class_list in cls.module_class.items():
            exed = [x for x in class_list if x not in exclude]
            if len(exed) == 0:
                continue
            msg = 'importing %s from %s' % (str(class_list), module_name)
            if verbose:
                logger.info(msg)
            else:
                logger.debug(msg)
            try:
                #m = importlib.__import__(module_name, globals(), locals(), class_list)
                m = importlib.import_module(module_name)
            except SelectiveMetaFinder.ExcludedModule as e:
                msg = 'Did not import %s. %s' % (str(class_list), str(e))
                if verbose:
                    logger.info(msg)
                else:
                    logger.debug(msg)
                #ety, enm, tb = sys.exc_info()
            except SyntaxError as e:
                msg = 'Could not import %s. %s' % (str(class_list), str(e))
                logger.error(msg)
                raise
            except ModuleNotFoundError as e:
                msg = 'Could not import %s. %s' % (str(class_list), str(e))
                if verbose:
                    logger.info(msg)
                else:
                    logger.debug(msg)
            else:
                for n in exed:
                    cls._package[n] = getattr(m, n)

        return

    def reloadClasses(cls):
        """ re-import classes in list. """
        for n, t in cls._classes.items():
            mo = importlib.import_module(t.__module__)
            importlib.reload(mo)
            m = importlib.__import__(t.__module__, globals(), locals(), [n])
            cls._classes[n] = getattr(m, n)

    # https://stackoverflow.com/a/1800999
    @property
    def mapping(cls):
        """ Returns the dictionary of classes allowed for deserialization, including the fdi built-ins and user added classes.

        Will update the classes if the list is empty
        """
        if len(cls._classes) == 0:
            cls.updateMapping()
        return cls._classes

    @mapping.setter
    def mapping(cls, c):
        """ Delegated to cls.update...().
        """
        raise NotImplementedError('Use Classes.updateMapping(c).')
        cls.updateMapping(c)


class Classes(metaclass=Classes_meta):
    """ A dictionary of class names and their class objects that are allowed to be deserialized.
    A fdi package built-in dictionary (in the format of locals() output) is kept internally.
    Users who need add more deserializable class can for example:
    class Myclass():
        ...
    Classes.classes.update({myClasses
    """

    pass


# globals()
# pdb.set_trace()
# Classes.importModuleClasses()



# -*- coding: utf-8 -*-
from .urn import Urn
from fdi.dataset.odict import ODict
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Taggable(object):
    """
    Definition of services provided by a product storage supporting versioning.
    """

    def __init__(self, **kwds):
        super(Taggable, self).__init__(**kwds)
        # {tag->{'urns':[urn]}
        self._tags = ODict()
        # {urn->{'tags':[tag], 'meta':meta}}
        self._urns = ODict()

    def getTags(self, urn=None):
        """ 
        Get all of the tags that map to a given URN.
        Get all known tags if urn is not specified.
        mh: returns an iterator.
        """
        if urn is None:
            return self._tags.keys()
        #uobj = Urn(urn=urn)
        return self._urns[urn]['tags']

    def getTagUrnMap(self):
        """
        Get the full tag->urn mappings.
        mh: returns an iterator
        """
        return zip(self._tags.keys(), map(lambda v: v['urns'], self._value()))

    def getUrn(self, tag):
        """
        Gets the URNs corresponding to the given tag. Returns an empty list if tag does not exist.
        """
        if tag not in self._tags:
            return []
        return self._tags[tag]['urns']

    def getUrnObject(self, tag):
        """
        Gets the URNobjects corresponding to the given tag.
        """
        l = [Urn(x) for x in self._tags[tag]['urns']]
        return l

    def removekey(self, key, themap, thename, othermap, othername):
        """
        Remove the given key.
        """
        vals = themap.pop(key)
        # remove all items whose v is key in the otherosit map
        for val in vals[othername]:
            othermap[val][thename].remove(key)
            if len(othermap[val][thename]) == 0:
                othermap[val].pop(thename)
                if len(othermap[val]) == 0:
                    othermap.pop(val)

    def removeTag(self, tag):
        """
        Remove the given tag from the tag and urn maps.
        """
        self.removekey(tag, self._tags, 'tags', self._urns, 'urns')

    def removeUrn(self, urn):
        """
        Remove the given urn from the tag and urn maps.
        """
        u = urn.urn if issubclass(urn.__class__, Urn) else urn
        self.removekey(u, self._urns, 'urns', self._tags, 'tags')

    def setTag(self, tag,  urn):
        """
        Sets the specified tag to the given URN.
        """
        u = urn.urn if issubclass(urn.__class__, Urn) else urn
        if u not in self._urns:
            raise ValueError(urn + ' not found in pool')
        else:
            self._urns[urn]['tags'].append(tag)

        if tag in self._tags:
            self._tags[tag]['urns'].append(u)
        else:
            self._tags[tag] = ODict(urns=[u])

    def tagExists(self, tag):
        """
        Tests if a tag exists.
        """
        return tag in self._tags



# -*- coding: utf-8 -*-
from ..dataset.odict import ODict
from ..dataset.serializable import serialize
from ..dataset.classes import Classes
from .urn import Urn, parseUrn, parse_poolurl
from .versionable import Versionable
from .taggable import Taggable
from .definable import Definable
from ..utils.common import pathjoin, fullname
from .productref import ProductRef
from .query import AbstractQuery, MetaQuery

import logging
import filelock
from copy import deepcopy
import os
import sys
from functools import lru_cache

import pdb

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    from urllib.parse import urlparse
else:
    PY3 = False
    from urlparse import urlparse

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


lockpathbase = '/tmp/fdi_locks'  # + getpass.getuser()
# lock time-out
locktout = 10


class ProductPool(Definable, Taggable, Versionable):
    """ A mechanism that can store and retrieve Products.

A product pool should not be used directly by users. The general user should access data in a ProductPool through a ProductStorage instance.

When implementing a ProductPool, the following rules need to be applied:

    1. Pools must guarantee that a Product saved via the pool saveProduct(Product) method is stored persistently, and that method returns a unique identifier (URN). If it is not possible to save a Product, an IOException shall be raised.
    2. A saved Product can be retrieved using the loadProduct(Urn) method, using as the argument the same URN that assigned to that Product in the earlier saveProduct(Product) call. No other Product shall be retrievable by that same URN. If this is not possible, an IOException or GeneralSecurityException is raised.
    3. Pools should not implement functionality currently implemented in the core package. Specifically, it should not address functionality provided in the Context abstract class, and it should not implement versioning/cloning support.

    """

    def __init__(self, poolname=None, poolurl=None, **kwds):
        """
        Creates and initializes a productpool.

        * poolname: if provided will override that in poolurl.
        * poolurl: needed to initialize.

        """
        super(ProductPool, self).__init__(**kwds)

        self._poolname = poolname

        self._poolpath, self._scheme, self._place, nm = \
            parse_poolurl(poolurl)
        self._poolurl = poolurl
        # self._pathurl = pr.netloc + pr.path
        # self._pathurl = None
        # {type|classname -> {'sn:[sn]'}}
        self._classes = ODict()

    def lockpath(self, op='w'):
        """ returns the appropriate path.

        creats the path if non-existing. Set lockpath-base permission to all-modify so other fdi users can use.
        op: 'r' for readlock no-reading) 'w' for writelock (no-writing)
        """
        if not os.path.exists(lockpathbase):
            os.makedirs(lockpathbase, mode=0o777)

        #from .httppool import HttpPool
        #from .httpclientpool import HttpClientPool

        p = self.transformpath(self._poolname)
        lp = pathjoin(lockpathbase, p.replace('/', '_'))

        if 1:
            return lp+'.read' if op == 'r' else lp+'.write'
        else:
            if not os.path.exists(lp):
                os.makedirs(lp)
                lf = pathjoin(lp, 'lock')
            return lf+'.read' if op == 'r' else lf+'.write'

    @lru_cache(maxsize=5)
    def transformpath(self, path):
        """ override this to changes the output from the input one (default) to something else.

        """
        base = self._poolpath
        if base != '':
            if path[0] == '/':
                path = base + path
            else:
                path = base + '/' + path
        return path

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern."""
        visitor.visit(self)

    def dereference(self, ref):
        """
        Decrement the reference count of a ProductRef.
        """
        self._urns[ref.urn]['refcnt'] -= 1

    def exists(self, urn):
        """
        Determines the existence of a product with specified URN.
        """
        return urn in self._urns

    def getDefinition(self):
        """
        Returns pool definition info which contains pool type and other pool specific configuration parameters
        """
        return super(ProductPool, self).getDefinition()

    def getId(self):
        """
        Gets the identifier of this pool.
        """
        return self._poolname

    def getPoolurl(self):
        """
        Gets the pool URL of this pool.
        """
        return self._poolurl

    def getPlace(self):
        """
        Gets the place of this pool.
        """
        return self._place

    def getProductClasses(self):
        """
        Returns all Product classes found in this pool.
        mh: returns an iterator.
        """
        return self._classes.keys()

    def getReferenceCount(self, ref):
        """
        Returns the reference count of a ProductRef.
        """
        return self._urns[ref.urn]['refcnt']

    def getScheme(self):
        """
        Gets the scheme of this pool.
        """
        return self._scheme

    def getUrnId(self):
        """
        Get the identifier of this pool used to build URN, usually it's same as id returned by getId().
        """
        return self.getId()

    def isAlive(self):
        """
        Test if the pool is capable of responding to commands.
        """
        return True

    def isEmpty(self):
        """
        Determines if the pool is empty.
        """
        return len(self._urns) == 0

    def loadDescriptors(self, urn):
        """
        Loads the descriptors belonging to specified URN.
        """
        return self._urns[urn]

    def schematicLoadProduct(self, resourcetype, index, serialized):
        """ to be implemented by subclasses to do the scheme-specific loading
        """
        raise(NotImplementedError)

    def loadProduct(self, urn, serialized=False):
        """
        Loads a Product belonging to specified URN.

        serialized: if True returns contents in serialized form.
        """
        poolname, resource, index = parseUrn(urn)
        if poolname != self._poolname:
            raise(ValueError('wrong pool: ' + poolname +
                             ' . This is ' + self._poolname))

        with filelock.FileLock(self.lockpath('w')):
            ret = self.schematicLoadProduct(
                resource, index, serialized=serialized)
        return ret

    def meta(self,  urn):
        """
        Loads the meta-data belonging to the product of specified URN.
        """
        return self._urns[urn]['meta']

    def reference(self,  ref):
        """
        Increment the reference count of a ProductRef.
        """
        self._urns[ref.urn]['refcnt'] += 1

    def schematicRemove(self, resourcetype, index):
        """ to be implemented by subclasses to do the scheme-specific removing
        """
        raise(NotImplementedError)

    def remove(self,  urn):
        """
        Removes a Product belonging to specified URN.
        """

        poolname, resource, index = parseUrn(urn)

        if self._poolname != poolname:
            raise ValueError(
                urn + ' is not from the pool ' + self._poolname)

        prod = resource
        sn = index

        self._classes, self._tags, self._urns = self.readHK()
        c, t, u = self._classes, self._tags, self._urns

        # save a copy for rolling back
        cs, ts, us = deepcopy(c), deepcopy(t), deepcopy(u)

        if urn not in u:
            raise ValueError(
                '%s not found in pool %s.' % (urn, self.getId()))

        with filelock.FileLock(self.lockpath('w')),\
                filelock.FileLock(self.lockpath('r')):
            self.removeUrn(urn)
            c[prod]['sn'].remove(sn)
            if len(c[prod]['sn']) == 0:
                del c[prod]
            try:
                res = self.schematicRemove(resourcetype=prod,
                                           index=sn)
            except Exception as e:
                msg = 'product ' + urn + ' removal failed'
                logger.debug(msg)
                # undo changes
                c, t, u = cs, ts, us
                raise e
        return res

    def removeAll(self):
        """
        Remove all pool data (self, products) and all pool meta data (self, descriptors, indices, etc.).
        """

        with filelock.FileLock(self.lockpath('w')),\
                filelock.FileLock(self.lockpath('r')):
            try:
                self.schematicWipe()
                self._classes.clear()
                self._tags.clear()
                self._urns.clear()
            except Exception as e:
                msg = self.getId() + 'wiping failed'
                logger.debug(msg)
                raise e
        logger.debug('Done.')

    def saveDescriptors(self,  urn,  desc):
        """
        Save/Update descriptors in pool.
        """

    def schematicSave(self, resourcetype, index, data, tag=None):
        """ to be implemented by subclasses to do the scheme-specific saving
        """
        raise(NotImplementedError)

    def saveProduct(self,  product, tag=None, geturnobjs=False, serialized=False):
        """
        Saves specified product and returns the designated ProductRefs or URNs.
        Saves a product or a list of products to the pool, possibly under the
        supplied tag, and return the reference (or a list of references is
        the input is a list of products), or Urns if geturnobjs is True.

        See pal document for pool structure.

        serialized: if True returns contents in serialized form.
        """
        c, t, u = self._classes, self._tags, self._urns
        # save a copy
        cs, ts, us = deepcopy(c), deepcopy(t), deepcopy(u)

        if not issubclass(product.__class__, list):
            prds = [product]
        else:
            prds = product
        res = []
        for prd in prds:
            pn = fullname(prd)
            with filelock.FileLock(self.lockpath('w')),\
                    filelock.FileLock(self.lockpath('r')):
                if pn in c:
                    sn = (c[pn]['currentSN'] + 1)
                else:
                    sn = 0
                    c[pn] = ODict(sn=[])

                c[pn]['currentSN'] = sn
                c[pn]['sn'].append(sn)

                urnobj = Urn(cls=prd.__class__,
                             poolname=self._poolname, index=sn)
                urn = urnobj.urn

                if urn not in u:
                    u[urn] = ODict(tags=[], meta=prd.meta)

                if tag is not None:
                    self.setTag(tag, urn)

                try:
                    self.schematicSave(resourcetype=pn,
                                       index=sn,
                                       data=prd,  tag=tag)
                except Exception as e:
                    msg = 'product ' + urn + ' saving failed'
                    logger.debug(msg)
                    # undo changes
                    c, t, u = cs, ts, us
                    raise e
            if geturnobjs:
                if serialized:
                    res.append(urn)
                else:
                    res.append(urnobj)
            else:
                rf = ProductRef(urn=urnobj)
                # it seems that there is no better way to set meta
                rf._meta = prd.getMeta()
                rf._poolname = self._poolname
                res.append(rf)
        logger.debug('generated ' + 'Urns ' if geturnobjs else 'prodRefs ' +
                     str(len(res)))
        if issubclass(product.__class__, list):
            return serialize(res) if serialized else res
        else:
            return serialize(res[0]) if serialized else res[0]

    def mfilter(self, q, cls=None, reflist=None, urnlist=None, snlist=None):
        """ returns filtered collection using the query.

        q is a MetaQuery
        valid inputs: cls and ns list; productref list; urn list
        """

        ret = []
        u = self._urns
        qw = q.getWhere()

        if reflist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for ref in reflist:
                    refmet = ref.getMeta()
                    m = refmet if refmet else u[ref.urn]['meta']
                    if eval(code):
                        ret.append(ref)
                return ret
            else:
                for ref in reflist:
                    refmet = ref.getMeta()
                    m = refmet if refmet else u[ref.urn]['meta']
                    if qw(m):
                        ret.append(ref)
                return ret
        elif urnlist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for urn in urnlist:
                    m = u[urn]['meta']
                    if eval(code):
                        ret.append(ProductRef(urn=urn, meta=m))
                return ret
            else:
                for urn in urnlist:
                    m = u[urn]['meta']
                    if qw(m):
                        ret.append(ProductRef(urn=urn, meta=m))
                return ret
        elif snlist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for n in snlist:
                    urno = Urn(cls=cls, poolname=self._poolname, index=n)
                    m = u[urno.urn]['meta']
                    if eval(code):
                        ret.append(ProductRef(urn=urno, meta=m))
                return ret
            else:
                for n in snlist:
                    urno = Urn(cls=cls, poolname=self._poolname, index=n)
                    m = u[urno.urn]['meta']
                    if qw(m):
                        ret.append(ProductRef(urn=urno, meta=m))
                return ret
        else:
            raise('Must give a list of ProductRef or urn or sn')

    def pfilter(self, q, cls=None, reflist=None, urnlist=None, snlist=None):
        """ returns filtered collection using the query.

        q is a AbstractQuery.
        valid inputs: cls and ns list; productref list; urn list
        """

        ret = []
        glbs = globals()
        u = self._urns
        qw = q.getWhere()
        var = q.getVariable()
        if var in glbs:
            savevar = glbs[var]
        else:
            savevar = 'not in glbs'

        if reflist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for ref in reflist:
                    glbs[var] = pref.getProduct()
                    if eval(code):
                        ret.append(ref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
            else:
                for ref in reflist:
                    glbs[var] = pref.getProduct()
                    if qw(m):
                        ret.append(ref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
        elif urnlist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for urn in urnlist:
                    pref = ProductRef(urn=urn)
                    glbs[var] = pref.getProduct()
                    if eval(code):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
            else:
                for urn in urnlist:
                    pref = ProductRef(urn=urn)
                    glbs[var] = pref.getProduct()
                    if qw(glbs[var]):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
        elif snlist:
            if isinstance(qw, str):
                code = compile(qw, 'py', 'eval')
                for n in snlist:
                    urno = Urn(cls=cls, poolname=self._poolname, index=n)
                    pref = ProductRef(urn=urno)
                    glbs[var] = pref.getProduct()
                    if eval(code):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
            else:
                for n in snlist:
                    urno = Urn(cls=cls, poolname=self._poolname, index=n)
                    pref = ProductRef(urn=urno)
                    glbs[var] = pref.getProduct()
                    if qw(glbs[var]):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
        else:
            raise('Must give a list of ProductRef or urn or sn')

    def select(self,  query, results=None):
        """
        Returns a list of references to products that match the specified query.
        """
        isMQ = issubclass(query.__class__, MetaQuery)
        isAQ = issubclass(query.__class__, AbstractQuery)
        if not isMQ and not isAQ:
            raise TypeError('not a Query')
        lgb = Classes.mapping
        t, v, w, a = query.getType(), query.getVariable(
        ), query.getWhere(), query.retrieveAllVersions()
        ret = []
        if results:
            this = (x for x in results if x.urnobj.getPoolId()
                    == self._poolname)
            if isMQ:
                ret += self.mfilter(q=query, reflist=this)
            else:
                ret += self.pfilter(q=query, reflist=this)
        else:
            for cname in self._classes:
                cls = lgb[cname.split('.')[-1]]
                if issubclass(cls, t):
                    if isMQ:
                        ret += self.mfilter(q=query, cls=cls,
                                            snlist=self._classes[cname]['sn'])
                    else:
                        ret += self.pfilter(q=query, cls=cls,
                                            snlist=self._classes[cname]['sn'])

        return ret

    def __repr__(self):
        return self.__class__.__name__ + ' { pool= ' + str(self._poolname) + ' }'



# -*- coding: utf-8 -*-
from .productpool import ProductPool
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MemPool(ProductPool):
    """ the pool will save all products in memory.
    """

    def __init__(self, **kwds):
        """ creates data structure if there isn't one. if there is, read and populate house-keeping records. create persistent files if not exist.
        """

        super(MemPool, self).__init__(**kwds)
        self._MemPool = {}
        # if self._poolname not in self._MemPool:
        #      self._MemPool[self._poolname] = {}
        c, t, u = self.readHK()

        logger.debug('created ' + self.__class__.__name__ +
                     ' ' + self._poolname + ' HK read.')

        self._classes.update(c)
        self._tags.update(t)
        self._urns.update(u)

    def getPoolSpace(self):
        """ returns the map of this memory pool.
        """
        return self._MemPool
        # if self._poolname in self._MemPool:
        #     return self._MemPool[self._poolname]
        # else:
        #     return None

    def readHK(self, hktype=None, serialized=False):
        """
        loads and returns the housekeeping data

        hktype: one of 'classes', 'tags', 'urns' to return. default is None to return alldirs
        serialized: if True return serialized form. Default is false.
        """

        if serialized:
            raise NotImplementedError
        if hktype is None:
            hks = ['classes', 'tags', 'urns']
        else:
            hks = [hktype]
        hk = {}
        myspace = self.getPoolSpace()
        for hkdata in hks:
            if len(myspace) == 0:
                r = {}
            else:
                r = myspace[hkdata]
            hk[hkdata] = r
        logger.debug('HK read from ' + self._poolname)
        return (hk['classes'], hk['tags'], hk['urns']) if hktype is None else hk[hktype]

    def writeHK(self):
        """
           save the housekeeping data to mempool
        """

        myspace = self.getPoolSpace()
        myspace['classes'] = self._classes
        myspace['tags'] = self._tags
        myspace['urns'] = self._urns

    def schematicSave(self, resourcetype, index, data, tag=None):
        """ 
        does the media-specific saving
        """
        resourcep = resourcetype + '_' + str(index)
        myspace = self.getPoolSpace()
        myspace[resourcep] = data
        self.writeHK()
        logger.debug('HK written')

    def schematicLoadProduct(self, resourcetype, index, serialized=False):
        """
        does the scheme-specific part of loadProduct.
        note that the index is given as a string.
        """
        if serialized:
            raise NotImplementedError
        indexstr = str(index)
        resourcep = resourcetype + '_' + indexstr
        myspace = self.getPoolSpace()
        return myspace[resourcep]

    def schematicRemove(self, resourcetype, index):
        """
        does the scheme-specific part of removal.
        """
        resourcep = resourcetype + '_' + str(index)
        myspace = self.getPoolSpace()
        del myspace[resourcep]
        self.writeHK()

    def schematicWipe(self):
        """
        does the scheme-specific remove-all
        """

        # logger.debug()
        p = self.getPoolSpace()
        p.clear()

        # del p will only delete p in current namespace, not anything in _MemPool
        # this wipes all mempools
        # pools = [x for x in self._MemPool]
        # for x in pools:
        #    del self._MemPool[x]
        # if self._poolname in self._MemPool:
        #    del self._MemPool[self._poolname]

    def getHead(self, ref):
        """ Returns the latest version of a given product, belonging
        to the first pool where the same track id is found.
        """



import copy

#      Dictionary of Commonly Used FITS Keywords
#
# This data dictionary contains FITS keywords that have been widely used
# within the astronomical community.  It is recommended that these
# keywords only be used as defined here.
#
# Resource: http://heasarc.gsfc.nasa.gov/docs/fcg/common_dict.html
#
FITS_keywords_HEASARC = {
    # -- HIFI, see HCSS-18735',
    'BAND': 'Band',
    # -- HERSCHEL-HSC-DOC-0662 Herschel Pointing Product Specification',
    'RADESYS': 'raDecSys',
    'AIRMASS': 'airMass',
    'APERTURE': 'aperture',
    'CHECKSUM': 'checksum',
    'CHECKVER': 'checksumVersion',
    'CONFIGUR': 'softwareConfiguration',
    'CDELT1': 'cdelt1',
    'CDELT2': 'cdelt2',
    'CDELT3': 'cdelt3',
    'CDELT4': 'cdelt4',
    'CROTA1': 'crota1',
    'CROTA2': 'crota2',
    'CROTA3': 'crota3',
    'CROTA4': 'crota4',
    'CRPIX1': 'crpix1',
    'CRPIX2': 'crpix2',
    'CRPIX3': 'crpix3',
    'CRPIX4': 'crpix4',
    'CRVAL1': 'crval1',
    'CRVAL2': 'crval2',
    'CRVAL3': 'crval3',
    'CRVAL4': 'crval4',
    'CTYPE1': 'ctype1',
    'CTYPE2': 'ctype2',
    'CTYPE3': 'ctype3',
    'CTYPE4': 'ctype4',
    'DATAMODE': 'preProcessorDataMode',
    'DATASUM': 'checksumData',
    'DATE-END': 'endDate',
    'DEC': 'declination',
    'DEC_NOM': 'declinationNominal',
    'DEC_OBJ': 'declinationObserved',
    'DEC_PNT': 'declinationPointing',
    'DEC_SCX': 'declinationSpacecraftX',
    'DEC_SCY': 'declinationSpacecraftY',
    'DEC_SCZ': 'declinationSpacecraftZ',
    'DETNAM': 'detectorName',
    'ELAPTIME': 'observationDuration',
    'EXPOSURE': 'exposureTime',
    'EXPTIME': 'exposureTime',
    'FILENAME': 'fileName',
    'FILETYPE': 'fileType',
    'FILTER': 'filter',
    # -- FILTERn': '',
    'GRATING': 'grating',
    # -- GRATINGn': '',
    # -- GRPIDn': '',
    # -- GRPLCn': '',
    # -- GRPNAME': '',
    # -- HDUCLASS': '',
    # -- HDUCLASn': '',
    # -- HDUDOC': '',
    # -- HDULEVEL': '',
    # -- HDUNAME': '',
    # -- HDUVER': '',
    # -- HDUVERS': '',
    # -- HIERARCH': '',
    # -- INHERIT': '',
    'LATITUDE': 'geographicLatitude',
    'LIVETIME': 'correctedExposureTime',
    'MOONANGL': 'moonObservationAngle',
    # -- NEXTEND': '',
    'OBJNAME': 'objectIauName',
    'OBS_ID': 'observationId',
    'OBS_MODE': 'observationInstrumentMode',
    'ONTIME': 'observationIntegrationTime',
    'ORIENTAT': 'positionAngleAxisY',
    'PA_PNT': 'positionAnglePointing',
    'PROGRAM': 'softwareTaskName',
    'RA': 'rightAscension',
    'RA_NOM': 'rightAscensionNominal',
    'RA_OBJ': 'rightAscensionObserved',
    'RA_PNT': 'rightAscensionPointing',
    'RA_SCX': 'rightAscensionSpacecraftX',
    'RA_SCY': 'rightAscensionSpacecraftY',
    'RA_SCZ': 'rightAscensionSpacecraftZ',
    'ROOTNAME': 'fileRootName',
    'SATURATE': 'dataSaturationValue',
    'SUNANGLE': 'sunObservationAngle',
    # -- TDBINn': '',
    # -- TDMAXn': '',
    # -- TDMINn': '',
    'TELAPSE': 'duration',
    'TIME-END': 'endTime',
    'TIME-OBS': 'startTime',
    'TITLE': 'title',
    # -- TLMAXn': '',
    # -- TLMINn': '',
    # -- TSORTKEY': '',
    # -- WAVELN': '',
    'WAVELNTH': 'wavelnth',
    'WAVEUNIT': 'waveunit',
}

# FITS keywords that are part of the FITS Standard definition:
#   Definition of the Flexible Image Transport System (FITS}
#   March 29, 1999
#   NOST 100-2.0
#   http://archive.stsci.edu/fits/fits_standard/
#
FITS_keywords_Standard = {
    'AUTHOR': 'author',
    'CREATOR': 'creator',
    'DATE-OBS': 'startDate',
    'DATE': 'creationDate',
    'EPOCH': 'epoch',
    'EQUINOX': 'equinox',
    'INSTRUME': 'instrument',
    'OBJECT': 'object',
    'OBSERVER': 'observer',
    'REFERENC': 'reference',
    'TELESCOP': 'telescope',
}

# Dictionary of commonly used CLASS keywords
#
FITS_keywords_CLASS = {
    'MODELNAM': 'modelName',
    'TEMPERAT': 'temperature',
    'MAXIS': 'MAXIS',
    'MAXIS1': 'MAXIS1',
    'MAXIS2': 'MAXIS2',
    'MAXIS3': 'MAXIS3',
    'MAXIS4': 'MAXIS4',
    'CTYPE1': 'CTYPE1',
    'CRVAL1': 'CRVAL1',
    'CDELT1': 'CDELT1',
    'CRPIX1': 'CRPIX1',
    'CTYPE2': 'CTYPE2',
    'CRVAL2': 'CRVAL2',
    'CDELT2': 'CDELT2',
    'CRPIX2': 'CRPIX2',
    'CTYPE3': 'CTYPE3',
    'CRVAL3': 'CRVAL3',
    'CDELT3': 'CDELT3',
    'CRPIX3': 'CRPIX3',
    'CTYPE4': 'CTYPE4',
    'CRVAL4': 'CRVAL4',
    'CDELT4': 'CDELT4',
    'CRPIX4': 'CRPIX4',
    'BANDWID': 'BANDWID',
    'RESTFREQ': 'RESTFREQ',
    'IMAGFREQ': 'IMAGFREQ',
    'LOFREQ': 'LOFREQ',
    'VELOCITY': 'VELOCITY',
    'DELTAV': 'DELTAV',
    'BLANK': 'BLANK',
    'OBSID': 'OBSID',
    'APID': 'APID',
    'BBID': 'BBID',
    'BBTYPE': 'BBTYPE',
    'BBNUMBER': 'BBNUMBER',
    'SEQNUMBE': 'SEQNUMBE',
    'BEAMEFF': 'BEAMEFF',
    'FORWEFF': 'FORWEFF',
    'APEREFF': 'APEREFF',
    'ETAL': 'ETAL',
    'ETAFSS': 'ETAFSS',
    'ANTGAIN': 'ANTGAIN',
    'BMAJ': 'BMAJ',
    'BMIN': 'BMIN',
    'BPA': 'BPA',
    'GAINIMAG': 'GAINIMAG',
    'TAU': 'TAU',
    'TAUIMAGE': 'TAUIMAGE',
    'TAUZENIT': 'TAUZENIT',
    'MH2O': 'MH2O',
    'HUMIDITY': 'HUMIDITY',
    'DEWPOINT': 'DEWPOINT',
    'PRESSURE': 'PRESSURE',
    'TOUTSIDE': 'TOUTSIDE',
    'WINDSPEE': 'WINDSPEE',
    'WINDDIRE': 'WINDDIRE',
    'SCAN': 'SCAN',
    'SUBSCAN': 'SUBSCAN',
    'TSYS': 'TSYS',
    'OBSTIME': 'OBSTIME',
    'EXPOSURE': 'EXPOSURE',
    'DATE-OBS': 'DATE-OBS',
    'DATE-RED': 'DATE-RED',
    'OBJECT': 'OBJECT',
    'LINE': 'LINE',
    'MOLECULE': 'MOLECULE',
    'TRANSITI': 'TRANSITI',
    'TELESCOP': 'TELESCOP',
    'NPHASE': 'NPHASE',
    'DELTAF1': 'DELTAF1',
    'PTIME1': 'PTIME1',
    'WEIGHT1': 'WEIGHT1',
    'DELTAF2': 'DELTAF2',
    'PTIME2': 'PTIME2',
    'WEIGHT2': 'WEIGHT2',
    'THOT': 'THOT',
    'TCOLD': 'TCOLD',
    'OBSERVER': 'OBSERVER',
    'PROJID': 'PROJID',
    'OBSMODE': 'OBSMODE',
    'TEMPSCAL': 'TEMPSCAL',
    'TCAL': 'TCAL',
    'TRX': 'TRX',
    'TIMESYS': 'TIMESYS',
    'SITELONG': 'SITELONG',
    'SITELAT': 'SITELAT',
    'SITEELEV': 'SITEELEV',
    'DATAMIN': 'DATAMIN',
    'DATAMAX': 'DATAMAX'
}

FITS_keywords_HCSS = {
    # Dictionary of commonly used HCSS keywords
    #
    'ACMSMODE': 'acmsMode',
    'ACTION': 'action',
    'ACTIVSTR': 'activeStrId',
    'ANALYST': 'dataAnalyst',
    'AOR': 'aorLabel',
    'AOT': 'aot',
    'APID': 'apid',
    'AVERAGIN': 'averaging',
    'BAND': 'band',
    'BASEMOD': 'baselineModel',
    'BASEPAR': 'baselineParams',
    'BBCOUNT': 'bbCount',
    'BBID': 'bbid',
    'BBTNAME': 'bbTypeName',
    'BBTYPE': 'bbType',
    'BIASVOLT': 'biasVoltage',
    'BITPOS': 'bitPos',
    'CALFILE': 'calFileId',
    'CALFILEV': 'calFileVersion',
    'CALVERS': 'calVersion',
    'CAMERA': 'camera',
    'CAMMODEL': 'cameraModel',
    'CD1_1': 'cd1_1',
    'CD1_2': 'cd1_2',
    'CD1_3': 'cd1_3',
    'CD2_1': 'cd2_1',
    'CD2_2': 'cd2_2',
    'CD2_3': 'cd2_3',
    'CD3_1': 'cd3_1',
    'CD3_2': 'cd3_2',
    'CD3_3': 'cd3_3',
    'CHANGLOG': 'changelog',
    'CHOPPLAT': 'chopperPlateau',
    'CH_STAT': 'status',
    'CONVELF': 'constVelFlag',
    'CONVFACT': 'conversionFactor',
    'CUSMODE': 'cusMode',
    'DEC': 'dec',
    'DEC_NOM': 'decNominal',
    'DEC_OBJ': 'decObject',
    'DELTAPIX': 'deltaPix',
    'DESC': 'description',
    'DETECTOR': 'arrayName',
    'END_WL': 'endWavelength',
    'ERROR': 'error',
    'EXP_TEXT': 'explanatoryText',
    'FINETIME': 'fineTime',
    'FORMATV': 'formatVersion',
    'GYROPQI': 'gyroPropQualIdx',
    'INSTMODE': 'instMode',
    'INTERPM': 'interpMethod',
    'JIGGLEID': 'jiggleId',
    'KEY_WAVE': 'keyWavelength',
    'LEVEL': 'level',
    'MAX_WAVE': 'maxWavelength',
    'MIN_WAVE': 'minWavelength',
    'MISSIONC': 'missionConfig',
    'MODELNAM': 'modelName',
    'NAIFID': 'naifId',
    'NCHOPCYC': 'numChopCyc',
    'NHIFSAA': 'numHifiSaa',
    'NJIGGPOS': 'numJigglePos',
    'NNODCYC': 'numNodCyc',
    'NODCYDEN': 'nodCycleNum',
    'NPACSSAA': 'numPacsSaa',
    'NRASTCOL': 'numRasterCol',
    'NRASTLIN': 'numRasterLines',
    'NSCANLIN': 'numScanLines',
    'NSPECTRA': 'numSpectra',
    'NSPIRESA': 'numSpireSaa',
    'OBJTYPE': 'objectType',
    'OBS_ID': 'obsid',
    'OBS_MODE': 'obsMode',
    'OBSSTATE': 'obsState',
    'ODNUMBER': 'odNumber',
    'OFF_POS': 'offPosFlag',
    'ONSRCTIM': 'onSourceTime',
    'ONTARF': 'onTargetFlag',
    'ORIGIN': 'origin',
    'OUTFIELD': 'outOfFieldFlag',
    'PC1_1': 'pc1_1',
    'PC1_2': 'pc1_2',
    'PC1_3': 'pc1_3',
    'PC2_1': 'pc2_1',
    'PC2_2': 'pc2_2',
    'PC2_3': 'pc2_3',
    'PC3_1': 'pc3_1',
    'PC3_2': 'pc3_2',
    'PC3_3': 'pc3_3',
    'PCAVEATS': 'proCaveats',
    'PIX_ROW': 'pixelRow',
    'PIX_STAT': 'status',
    'PMRA': 'pmRA',
    'PMDEC': 'pmDEC',
    'POINTMOD': 'pointingMode',
    'POSANGLE': 'posAngle',
    'PROCMODE': 'processingMode',
    'PRODNOTE': 'productNotes',
    'PROPOSAL': 'proposal',
    'Q_FLAG': 'qualityFlag',
    'RASTCOL': 'rasterColumnNum',
    'RASTLINE': 'rasterLineNum',
    'RA': 'ra',
    'RADESYS': 'raDeSys',
    'RA_ERR': 'raErr',
    'RA_NOM': 'raNominal',
    'RA_OBJ': 'raObject',
    'READOUTS': 'readouts',
    'REFEREN': 'references',
    'REFPIXEL': 'refPixel',
    'ROLL': 'roll',
    'SAA': 'saa',
    'SATURATE': 'saturation',
    'SAT_SIGN': 'satValuesSigned',
    'SAT_UNSG': 'satValuesUnsigned',
    'SCANLINE': 'scanLineNum',
    'SED_VER': 'sedVersion',
    'SERENDIP': 'serendipityFlag',
    'SIAM_ID': 'siamId',
    'SKY_RES': 'skyResolution',
    'SLEWFLAG': 'slewFlag',
    'SLEWTIME': 'slewTime',
    'SLICENUM': 'sliceNumber',
    'SOURCE': 'source',
    'SPEC_NUM': 'specNum',
    'SPEC_RES': 'spectralResolution',
    'SRC_DETC': 'sourceDetector',
    'SRC_SMEC': 'sourceSmec',
    'SRCFILE': 'sourceFile',
    'START_WL': 'startWavelength',
    'STATUS': 'state',
    'STR_I_ST': 'strInterlacingStatus',
    'STR_Q_ID': 'strQualIdx',
    'SUBINST': 'subinstrumentId',
    'SUBSYS': 'subsystem',
    'TEMPERAT': 'temperature',
    'THRESHOL': 'calThreshold',
    'TYPE': 'type',
    'VARIABLE': 'variability',
    'VELDEF': 'velocityDefinition',
    'VERSION': 'version',
    'VER_NOTE': 'versionNotes',
    'VFRAME': 'radialVelocity',
    'WAVE_ID': 'wavelengthId',
    'WAVEDESC': 'wavedescription',
    'WAVELNTH': 'wavelength',
    'WCS_REF': 'wcsReference',
    'WCS_TYPE': 'wcsType',
    'WHEELPOS': 'wheelPos',
    'ZERO_OFF': 'zeroPointOffset',
    # -- ',
    # -- clean FITS TODO move',
    # -- HCSS-18608 ',
    'CALBLKNO': 'blockNumber',
    'LINECENT': 'lWave',
    'LINES': 'lines',
    'QUERYTME': 'queryTime',
    'QURYTMST': 'queryTimeAsString',
    'RANGEREP': 'repeatRange',
    'RASSTEPL': 'lineStep',
    'WAVEDENS': 'density',
    # -- HCSS-18735',
    'LFREQMAX': 'obsFreqLsbMax',
    'LFREQMIN': 'obsFreqLsbMin',
    'LOFREQ': 'loFrequency',
    'LODOPPAV': 'loFreqAvg',
    'UFREQMIN': 'obsFreqUsbMin',
    'UFREQMAX': 'obsFreqUsbMax',
    # -- HCSS-19017',
    'SPECSYS': 'freqFrame',
    # -- HCSS-18514 HERSCHEL-HSC-DOC-0816 Herschel Auxiliary Products Specification',
    # -- Mission timeline summary product (auxMtls}',
    'PSFVER': 'psfVersion',
    'POSVER': 'posVersion',
    'EPOSVER': 'eposVersion',
    # -- Events log product data eventsLogData',
    'PKT_TYPE': 'packetType',
    'VC': 'virtualChannel',
    # -- Observation request data ObsRequestData',
    'OBSREQID': 'obsRequestId',
    'OBREQVER': 'obsRequestVersion',
    'TARGET': 'targetName',
    'TARGTYPE': 'targetType',
    'SUB_INST': 'subinstrument',
    'CUS_MODE': 'observingMode',
    'TIME_EST': 'timeEstimate',
    'OVERHEAD': 'overhead',
    'CHOP_ANG': 'chopperAngle',
    'CHOPAVO1': 'chopperAvoidAngle1',
    'CHOPAVO2': 'chopperAvoidAngle2',
    'MAPAVOI1': 'mapAvoidAngle1',
    'MAPAVOI2': 'mapAvoidAngle2',
    'TSLEWMIN': 'timeSlewMin',
    'TIH': 'timeInitialHold',
    'TFH': 'timeFinalHold',
    'APER_ID': 'apertureId',
    'Y_OFFSET': 'yOffset',
    'Z_OFFSET': 'zOffset',
    'FIX_FRAM': 'fixedFrame',
    'PATT': 'pattAngle',
    'TP': 'timePointing',
    'D1': 'patternD1',
    'D2': 'patternD2',
    'K_OFF': 'kOff',
    'TOP': 'timeOffPointing',
    'RA_OFF': 'raOff',
    'DEC_OFF': 'decOff',
    'SCANRATE': 'scanRate',
    'NUM_NODS': 'numberNods',
    'CHOPTHRO': 'chopperThrow',
    'PATTNOD': 'pattNodAngle',
    'TPA': 'timePointingA',
    'TPB': 'timePointingB',
    'TLOADMIN': 'timeLoadMin',
    'NLOAD': 'numLoad',
    'THOLD': 'timeHold',
    'NHOLD': 'numHold',
    'NRPEAT': 'numRepeat',
    'TREPMIN': 'timeRepeatMin',
    'NCYCLES': 'numCyckes',
    'XRPEAT': 'xRepeat',
    'PATTX': 'pattAngleX',
    'NUMLINEX': 'numLinesX',
    'D1X': 'patternD1X',
    'D2X': 'patternD2X',
    'NHOLDX': 'numHoldX',
    'START_B': 'startAtb',
    # -- Proposal data ProposalData',
    'PROPNUID': 'proposalNumId',
    'PROPVER': 'propVersion',
    'PROPCAT': 'propCategory',
    'PROPSCAT': 'propScienceCategory',
    'PROPTITL': 'propTitle',
    'PROP_PI': 'propPI',
    'PROPCOMM': 'proposerComment',
    'REQTIME': 'requestedTime',
    'TIMPRIO1': 'timeAllocPriority1',
    'TIMPRIO2': 'timeAllocPriority2',
    'TECHEVAL': 'technicalEvaluation',
    'TECEVALC': 'technicalEvalComment',
    # -- Monitoring Data from CCUA MonitorCCUA (and Monitoring Data from CCUB MonitorCCUB}',
    'RAW_SPID': 'rawPacketSpid',
    # -- SSO Horizons Ephemerides',
    'SSO_NAME': 'ssoName',
    'SSONAMSR': 'ssoNameSource',
    'CEN_NAM': 'centerBodyName',
    'CENSINAM': 'centerSiteName',
    'CENGLONG': 'centerGeodeticLong',
    'CENGLAT': 'centerGeodeticLat',
    'CENGALT': 'centerGeodeticAlt',
    'CENCLONG': 'centerCylindricLong',
    'CENCXY': 'centerCylindricXY',
    'CENCZ': 'centerCylindricZ',
    'CENRADII': 'centerRadii',
    'PERNASRC': 'perturberNameSource',
    'OUTUNITS': 'outputUnits',
    'REFFRAME': 'referenceFrame',
    'OUTTYPE': 'outputType',
    'COOSYSTM': 'coordinateSystem',
    'ELEMDATE': 'elementsDate',
    'ELEM_EC': 'elementEC',
    'ELEM_QR': 'elementQR',
    'ELEM_IN': 'elementIN',
    'ELEM_OM': 'elementOM',
    'ELEM_W': 'elementW',
    'ELEM_TP': 'elementTP',
    'ELEM_N': 'elementN',
    'ELEM_MA': 'elementMA',
    'ELEM_TA': 'elementTA',
    'ELEM_A': 'elementA',
    'ELEM_AD': 'elementAD',
    'ELEM_PER': 'elementPER',
    'SSO_RAD': 'ssoRadius',
    'SSO_GM': 'ssoGM',
    'AST_BV': 'asteroidColourBV',
    'AST_H': 'asteroidH',
    'AST_G': 'asteroidG',
    'ASTROTPE': 'asteroidRotPeriod',
    'AST_ALB': 'asteroridAlbedo',
    'ASTSTYPE': 'asteroidSpectralType',
    'COM_M1': 'cometM1',
    'COM_M2': 'cometM2',
    'COM_K1': 'cometK1',
    'COM_K2': 'cometK2',
    'COM_PHCO': 'cometPhaseCoeff',
    'COM_A1': 'cometA1',
    'COM_A2': 'cometA2',
    'COM_A3': 'cometA3',
    'COM_DT': 'cometDT',
    # -- ACMS TM Product (auxAcmsTM}',
    'G1RATBIA': 'Gyr1RateBias',
    'G1RATSCA': 'Gyr1RateScale',
    'G123OI11': 'Gyr123OrientInv11',
    'G123OI12': 'Gyr123OrientInv12',
    'G123OI13': 'Gyr123OrientInv13',
    'G123OI21': 'Gyr123OrientInv21',
    'G123OI22': 'Gyr123OrientInv22',
    'G123OI23': 'Gyr123OrientInv23',
    'G123OI31': 'Gyr123OrientInv31',
    'G123OI32': 'Gyr123OrientInv32',
    'G123OI33': 'Gyr123OrientInv33',
    'G124OI11': 'Gyr124OrientInv11',
    'G124OI12': 'Gyr124OrientInv12',
    'G124OI13': 'Gyr124OrientInv13',
    'G124OI21': 'Gyr124OrientInv21',
    'G124OI22': 'Gyr124OrientInv22',
    'G124OI23': 'Gyr124OrientInv23',
    'G124OI31': 'Gyr124OrientInv31',
    'G124OI32': 'Gyr124OrientInv32',
    'G124OI33': 'Gyr124OrientInv33',
    'G134OI11': 'Gyr134OrientInv11',
    'G134OI12': 'Gyr134OrientInv12',
    'G134OI13': 'Gyr134OrientInv13',
    'G134OI21': 'Gyr134OrientInv21',
    'G134OI22': 'Gyr134OrientInv22',
    'G134OI23': 'Gyr134OrientInv23',
    'G134OI31': 'Gyr134OrientInv31',
    'G134OI32': 'Gyr134OrientInv32',
    'G134OI33': 'Gyr134OrientInv33',
    'G2RATBIA': 'Gyr2RateBias',
    'G2RATSCA': 'Gyr2RateScale',
    'G234OI11': 'Gyr234OrientInv11',
    'G234OI12': 'Gyr234OrientInv12',
    'G234OI13': 'Gyr234OrientInv13',
    'G234OI21': 'Gyr234OrientInv21',
    'G234OI22': 'Gyr234OrientInv22',
    'G234OI23': 'Gyr234OrientInv23',
    'G234OI31': 'Gyr234OrientInv31',
    'G234OI32': 'Gyr234OrientInv32',
    'G234OI33': 'Gyr234OrientInv33',
    'G3RATBIA': 'Gyr3RateBias',
    'G3RATSCA': 'Gyr3RateScale',
    'G4RATBIA': 'Gyr4RateBias',
    'G4RATSCA': 'Gyr4RateScale',
    'SCMATM11': 'HScmAttMan_11',
    'SCMATM12': 'HScmAttMan_12',
    'SCMATM13': 'HScmAttMan_13',
    'SCMATM14': 'HScmAttMan_14',
    'SCMATM21': 'HScmAttMan_21',
    'SCMATM22': 'HScmAttMan_22',
    'SCMATM23': 'HScmAttMan_23',
    'SCMATM24': 'HScmAttMan_24',
    'SCMATM31': 'HScmAttMan_31',
    'SCMATM32': 'HScmAttMan_32',
    'SCMATM33': 'HScmAttMan_33',
    'SCMATM34': 'HScmAttMan_34',
    'SCMATM41': 'HScmAttMan_41',
    'SCMATM42': 'HScmAttMan_42',
    'SCMATM43': 'HScmAttMan_43',
    'SCMATM44': 'HScmAttMan_44',
    'SCMATP11': 'HScmAttPoi_11',
    'SCMATP12': 'HScmAttPoi_12',
    'SCMATP13': 'HScmAttPoi_13',
    'SCMATP14': 'HScmAttPoi_14',
    'SCMATP21': 'HScmAttPoi_21',
    'SCMATP22': 'HScmAttPoi_22',
    'SCMATP23': 'HScmAttPoi_23',
    'SCMATP24': 'HScmAttPoi_24',
    'SCMATP31': 'HScmAttPoi_31',
    'SCMATP32': 'HScmAttPoi_32',
    'SCMATP33': 'HScmAttPoi_33',
    'SCMATP34': 'HScmAttPoi_34',
    'SCMATP41': 'HScmAttPoi_41',
    'SCMATP42': 'HScmAttPoi_42',
    'SCMATP43': 'HScmAttPoi_43',
    'SCMATP44': 'HScmAttPoi_44',
    'SCDRMA11': 'HScmDrifMan_11',
    'SCDRMA12': 'HScmDrifMan_12',
    'SCDRMA13': 'HScmDrifMan_13',
    'SCDRMA14': 'HScmDrifMan_14',
    'SCDRMA21': 'HScmDrifMan_21',
    'SCDRMA22': 'HScmDrifMan_22',
    'SCDRMA23': 'HScmDrifMan_23',
    'SCDRMA24': 'HScmDrifMan_24',
    'SCDRMA31': 'HScmDrifMan_31',
    'SCDRMA32': 'HScmDrifMan_32',
    'SCDRMA33': 'HScmDrifMan_33',
    'SCDRMA34': 'HScmDrifMan_34',
    'SCDRPO11': 'HScmDrifPoin_11',
    'SCDRPO12': 'HScmDrifPoin_12',
    'SCDRPO13': 'HScmDrifPoin_13',
    'SCDRPO14': 'HScmDrifPoin_14',
    'SCDRPO21': 'HScmDrifPoin_21',
    'SCDRPO22': 'HScmDrifPoin_22',
    'SCDRPO23': 'HScmDrifPoin_23',
    'SCDRPO24': 'HScmDrifPoin_24',
    'SCDRPO31': 'HScmDrifPoin_31',
    'SCDRPO32': 'HScmDrifPoin_32',
    'SCDRPO33': 'HScmDrifPoin_33',
    'SCDRPO34': 'HScmDrifPoin_34',
    'SCDRC101': 'HScmDriftrck101',
    'SCDRC102': 'HScmDriftrck102',
    'SCDRC103': 'HScmDriftrck103',
    'SCDRC104': 'HScmDriftrck104',
    'SCDRC105': 'HScmDriftrck105',
    'SCDRC106': 'HScmDriftrck106',
    'SCDRC107': 'HScmDriftrck107',
    'SCDRC108': 'HScmDriftrck108',
    'SCDRC109': 'HScmDriftrck109',
    'SCDRC110': 'HScmDriftrck110',
    'SCDRC111': 'HScmDriftrck111',
    'SCDRC201': 'HScmDriftrck201',
    'SCDRC202': 'HScmDriftrck202',
    'SCDRC203': 'HScmDriftrck203',
    'SCDRC204': 'HScmDriftrck204',
    'SCDRC205': 'HScmDriftrck205',
    'SCDRC206': 'HScmDriftrck206',
    'SCDRC207': 'HScmDriftrck207',
    'SCDRC208': 'HScmDriftrck208',
    'SCDRC209': 'HScmDriftrck209',
    'SCDRC210': 'HScmDriftrck210',
    'SCDRC211': 'HScmDriftrck211',
    'SCDRC301': 'HScmDriftrck301',
    'SCDRC302': 'HScmDriftrck302',
    'SCDRC303': 'HScmDriftrck303',
    'SCDRC304': 'HScmDriftrck304',
    'SCDRC305': 'HScmDriftrck305',
    'SCDRC306': 'HScmDriftrck306',
    'SCDRC307': 'HScmDriftrck307',
    'SCDRC308': 'HScmDriftrck308',
    'SCDRC309': 'HScmDriftrck309',
    'SCDRC310': 'HScmDriftrck310',
    'SCDRC311': 'HScmDriftrck311',
    'GYMEADEL': 'HGyrMeasDelay',
    'STROUDEL': 'HStrOutputDelay',
    # -- Orbit info Common Metadata',
    'CENTNAME': 'centerName',
    # -- Orbit info TableDataset',
    'INTERPDG': 'interpDegree',
    # -- Pointing info TableDataset',
    'RSLINNUM': 'rasterLineNum',
    'RSCOLNUM': 'rasterColumnNum',
    'SCLINNUM': 'scanLineNum',
    'XSCANNUM': 'crossScanNum',
    'CUSPTNUM': 'customMapPointNum',
    'NOD_NUM': 'nodCycleNum',
    'ABPOSID': 'abPosId',
    'POINTID': 'pointModeId',
    'SEREN_FL': 'serendipityFlag',
    'STR_USE': 'strInUse',
    # -- HCSS-20112 Gyro Quality Indicators',
    'GYR_QUAL': 'gyroAttQuality',
    'PROBTHRE': 'probThreshold',
    'GYR_COV': 'gyroAttCoverage',
    'PROB_BAD': 'probBad',
    'COVTHRE': 'coverageThresh',
    'GYR_SUSP': 'gyroAttSuspicious',
    # -- SIAM product common metadata',
    'ACTSTRID': 'activeStrId',
    'NSPIRSAA': 'nSpireSaa',
    'NHIFISAA': 'nHifiSaa',
    # -- SIAM matrices - Array dataset',
    'DATE-VAL': 'validityStart',
    # -- Calibrated detectors accumulation data CalAccumData',
    'PROTONE1': 'protonE1',
    'PROTONE2': 'protonE2',
    'PROTONE3': 'protonE3',
    'PROTONE4': 'protonE4',
    'PROTONE': 'protonE5',
    'ELECT_E1': 'electronE1',
    'ELECT_E2': 'electronE2',
    'ELECT_E3': 'electronE3',
    'ELECT_E4': 'electronE4',
    # -- HCSS-18984 HIFI',
    'LOFRQMIN': 'loFreqMin',
    'LOFRQMAX': 'loFreqMax',
    'FREQMIN': 'obsFreqMin',
    'FREQMAX': 'obsFreqMax',
    'LOFRQSTA': 'loFrequencyStart',
    'LOFRQEND': 'loFrequencyEnd',
    # -- HCSS-19274 HIFI',
    'IS-WBSH': 'wbsHscience',
    'IS-WBSV': 'wbsVscience',
    'IS-HRSH': 'hrsHscience',
    'IS-HRSV': 'hrsVscience',
    'REDUNDCY': 'redundancy',
    'FREQ_THR': 'fsThrow',
    'NOISMAXU': 'noiseMaxUsb',
    'NOISMINU': 'noiseMinUsb',
    'NOISMAXL': 'noiseMaxLsb',
    'NOISMINL': 'noiseMinLsb',
    'NOISMIND': 'noiseDSBMin',
    'NOISMAXD': 'noiseDSBMax',
    'NOISMINS': 'noiseSSBMin',
    'NOISMAXS': 'noiseSSBMax',
    'SAAMEAN': 'solarAspectAngleMean',
    'SAARMS': 'solarAspectAngleRms',
    # -- HCSS-19347 HIFI',
    'BACKEND': 'backend',
    # -- HCSS-19350 HIFI',
    'NSMINWID': 'noiseMinWidth',
    'NSMAXWID': 'noiseMaxWidth',
    'TMBREFFQ': 'tmbReference',
    'NSREFFRQ': 'noiseRefFrequency',
    'TOTNSEFF': 'totNoiseEfficiency',
    'DRFTNSCT': 'driftNoiseContrib',
    # -- HCSS-19348',
    'ORBITFIL': 'orbitEphemerisSourceFile',
    # -- HCSS-19360',
    'SYSSRC': 'redshiftFrame',
    # -- HCSS-19449',
    'MIXER': 'mixer',
    'AGEOM': 'aGeom',
    # -- HCSS-19358',
    'TEMPSCAL': 'temperatureScale',
    # -- HCSS-19578',
    'ETAA': 'apertureEfficiency',
    'ETAL': 'forwardEff',
    'ETALCAL': 'forwardEfficiency',
    'ETAMB': 'beamEff',
    'ETAMBCAL': 'mainBeamEfficiency',
    'HPBW': 'hpbw',
    # -- HCSS-19682',
    'RMSMAXU': 'rmsMaxUsb',
    'RMSMINU': 'rmsMinUsb',
    'RMSMAXL': 'rmsMaxLsb',
    'RMSMINL': 'rmsMinLsb',
    'RMSMAXS': 'rmsSSBMax',
    'RMSMINS': 'rmsSSBMin',
    'RMSMAXD': 'rmsDSBMax',
    'RMSMIND': 'rmsDSBMin',
    # -- HCSS-19882',
    'RMSNATU': 'rmsNativeUsb',
    'RMSNATL': 'rmsNativeLsb',
    'RMSNATS': 'rmsSSBNative',
    'RMSNATD': 'rmsDSBNative',
    # -- HCSS-20388 HCSS-20389',
    'NSRATIHV': 'rmsNoiseHV',
    'NOISRATI': 'rmsNoise',
    # -- HCSS-20733',
    'WCSNAME': 'wcsName',
    # -- HCSS-20497',
    'LOTHRDOP': 'LoThrow',
    'LO_THR': 'LoThrow_measured',
    'MIXCURH': 'MJC_Hor',
    'MIXCURV': 'MJC_Ver',
    'OBSWPATC': 'OBS-patch',
    'OBSWREVI': 'OBS-revision',
    'OBSWVERS': 'OBS-version',
    # Note that BBNUMBER is in CLASS too',
    'BBNUMBER': 'bbnumber',
    'CBBTEMP': 'cbbTemp',
    'HBBTEMP': 'hbbTemp',
    'FREQRES': 'resolution_resampled',
    'CRDER1': 'raError',
    'CRDER2': 'decError',
    'TCRD1': 'longitudeError',
    'TCRD2': 'latitudeError',
    'LSBGAIN': 'lsbGain',
    'LSBGAIN0': 'lsbGain_0',
    'LSBGAIN1': 'lsbGain_1',
    'LSBGAIN2': 'lsbGain_2',
    'LSBGAIN3': 'lsbGain_3',
    'USBGAIN': 'usbGain',
    'USBGAIN0': 'usbGain_0',
    'USBGAIN1': 'usbGain_1',
    'USBGAIN2': 'usbGain_2',
    'USBGAIN3': 'usbGain_3',
    'INTEGTIM': 'integrationTime',
    'ISFOLDED': 'isFolded',
    'SIDEBAND': 'sideband',
    'SUBBAND': 'subband',
    'SUBLEN1': 'subbandlength_1',
    'SUBLEN2': 'subbandlength_2',
    'SUBLEN3': 'subbandlength_3',
    'SUBLEN4': 'subbandlength_4',
    'SUBSTA1': 'subbandstart_1',
    'SUBSTA2': 'subbandstart_2',
    'SUBSTA3': 'subbandstart_3',
    'SUBSTA4': 'subbandstart_4',
    'MEDTSYS': 'tsys_median',
    'HSOSSBVX': 'velocity_hso_1',
    'HSOSSBVY': 'velocity_hso_2',
    'HSOSSBVZ': 'velocity_hso_3',
    'POSANGER': 'posAngleError',
    # -- Quality flags mapping',
    'BADCOORD': 'qflag_BADCOORD_p',
    # -- Fri Jan 22 15:31:00 CET 2010',
    'ADCLATCH': 'qflag_ADCLATCH_p',
    'BSMMVSCN': 'qflag_BSMMVSCN_p',
    'CCPOSUNC': 'qflag_CCPOSUNC_p',
    'CCPOSUNV': 'qflag_CCPOSUNC_p_v',
    'BATHTEMP': 'qflag_BATHTEMP_p',
    'BATHTEMV': 'qflag_BATHTEMP_p_v',
    'SLWBATHT': 'qflag_SLWBATHT_p',
    'SLWBATHV': 'qflag_SLWBATHT_p_v',
    'SSWBATHT': 'qflag_SSWBATHT_p',
    'SSWBATHV': 'qflag_SSWBATHT_p_v',
    'SCALTEMP': 'qflag_SCALTEMP_p',
    'SCALTEMV': 'qflag_SCALTEMP_p_v',
    'CAL2TEMP': 'qflag_CAL2TEMP_p',
    'CAL2TEMV': 'qflag_CAL2TEMP_p_v',
    'CAL4TEMP': 'qflag_CAL4TEMP_p',
    'CAL4TEMV': 'qflag_CAL4TEMP_p_v',
    'TEL1TEMP': 'qflag_TEL1TEMP_p',
    'TEL1TEMV': 'qflag_TEL1TEMP_p_v',
    'TEL2TEMP': 'qflag_TEL2TEMP_p',
    'TEL2TEMV': 'qflag_TEL2TEMP_p_v',
    'TEL3TEMP': 'qflag_TEL3TEMP_p',
    'TEL3TEMV': 'qflag_TEL3TEMP_p_v',
    'DDROPPED': 'qflag_DDROPPED_p',
    'DDROPPEV': 'qflag_DDROPPED_p_v',
    'TLINESNR': 'qflag_TLINESNR_p',
    'TLINESNV': 'qflag_TLINESNR_p_v',
    'DFLUXPLW': 'qflag_DFLUXPLW_p',
    'DFLUXPLV': 'qflag_DFLUXPLW_p_v',
    'DFLUXPMW': 'qflag_DFLUXPMW_p',
    'DFLUXPMV': 'qflag_DFLUXPMW_p_v',
    'DFLUXPSW': 'qflag_DFLUXPSW_p',
    'DFLUXPSV': 'qflag_DFLUXPSW_p_v',
    'MPLWFAIL': 'qflag_MPLWFAIL_p',
    'MPMWFAIL': 'qflag_MPMWFAIL_p',
    'MPSWFAIL': 'qflag_MPSWFAIL_p',
    'MPLWMAXI': 'qflag_MPLWMAXI_p',
    'MPMWMAXI': 'qflag_MPMWMAXI_p',
    'MPSWMAXI': 'qflag_MPSWMAXI_p',
    'DLATMWLW': 'qflag_DLATMWLW_p',
    'DLATMWLV': 'qflag_DLATMWLW_p_v',
    'DLATSWLW': 'qflag_DLATSWLW_p',
    'DLATSWLV': 'qflag_DLATSWLW_p_v',
    'DLATSWMW': 'qflag_DLATSWMW_p',
    'DLATSWMV': 'qflag_DLATSWMW_p_v',
    'DLONMWLW': 'qflag_DLONMWLW_p',
    'DLONMWLV': 'qflag_DLONMWLW_p_v',
    'DLONSWLW': 'qflag_DLONSWLW_p',
    'DLONSWLV': 'qflag_DLONSWLW_p_v',
    'DLONSWMW': 'qflag_DLONSWMW_p',
    'DLONSWMV': 'qflag_DLONSWMW_p_v',
    'MFAILPLW': 'qflag_MFAILPLW_p',
    'MFAILPLV': 'qflag_MFAILPLW_p_v',
    'MFAILPMW': 'qflag_MFAILPMW_p',
    'MFAILPMV': 'qflag_MFAILPMW_p_v',
    'MFAILPSW': 'qflag_MFAILPSW_p',
    'MFAILPSV': 'qflag_MFAILPSW_p_v',
    'MISFRING': 'qflag_MISFRING_p',
    'MISFRINV': 'qflag_MISFRING_p_v',
    'NFITSPLW': 'qflag_NFITSPLW_p',
    'NFITSPMW': 'qflag_NFITSPMW_p',
    'NFITSPSW': 'qflag_NFITSPSW_p',
    'PLWVOLK3': 'qflag_PLWVOLK3_p',
    'PLWVOLKV': 'qflag_PLWVOLK3_p_v',
    'PMWVOLK3': 'qflag_PMWVOLK3_p',
    'PMWVOLKV': 'qflag_PMWVOLK3_p_v',
    'PSWVOLK3': 'qflag_PSWVOLK3_p',
    'PSWVOLKV': 'qflag_PSWVOLK3_p_v',
    'SLWVOLK3': 'qflag_SLWVOLK3_p',
    'SLWVOLKV': 'qflag_SLWVOLK3_p_v',
    'SSWVOLK3': 'qflag_SSWVOLK3_p',
    'SSWVOLKV': 'qflag_SSWVOLK3_p_v',
    'BSMCHPSL': 'qflag_BSMCHPSL_p',
    'BSMCHPSV': 'qflag_BSMCHPSL_p_v',
    'BSMJGGSL': 'qflag_BSMJGGSL_p',
    'BSMJGGSV': 'qflag_BSMJGGSL_p_v',
    'SLWCLIPN': 'qflag_SLWCLIPN_p',
    'SLWCLIPV': 'qflag_SLWCLIPN_p_v',
    'SSWCLIPN': 'qflag_SSWCLIPN_p',
    'SSWCLIPV': 'qflag_SSWCLIPN_p_v',
    'NMISSPOS': 'qflag_NMISSPOS_p',
    'NMISSPOV': 'qflag_NMISSPOS_p_v',
    'SLWGLITN': 'qflag_SLWGLITN_p',
    'SLWGLITV': 'qflag_SLWGLITN_p_v',
    'SSWGLITN': 'qflag_SSWGLITN_p',
    'SSWGLITV': 'qflag_SSWGLITN_p_v',
    'SLWPHASE': 'qflag_SLWPHASE_p',
    'SLWPHASV': 'qflag_SLWPHASE_p_v',
    'SSWPHASE': 'qflag_SSWPHASE_p',
    'SSWPHASV': 'qflag_SSWPHASE_p_v',
    'FDIFFPLW': 'qflag_FDIFFPLW_p',
    'FDIFFPLV': 'qflag_FDIFFPLW_p_v',
    'FDIFFPMW': 'qflag_FDIFFPMW_p',
    'FDIFFPMV': 'qflag_FDIFFPMW_p_v',
    'FDIFFPSW': 'qflag_FDIFFPSW_p',
    'FDIFFPSV': 'qflag_FDIFFPSW_p_v',
    'PLWNEGFL': 'qflag_PLWNEGFL_p',
    'PMWNEGFL': 'qflag_PMWNEGFL_p',
    'PSWNEGFL': 'qflag_PSWNEGFL_p',
    'DPOSPLWC': 'qflag_DPOSPLWC_p',
    'DPOSPLWV': 'qflag_DPOSPLWC_p_v',
    'DPOSPMWC': 'qflag_DPOSPMWC_p',
    'DPOSPMWV': 'qflag_DPOSPMWC_p_v',
    'DPOSPSWC': 'qflag_DPOSPSWC_p',
    'DPOSPSWV': 'qflag_DPOSPSWC_p_v',
    'DPOSMWLW': 'qflag_DPOSMWLW_p',
    'DPOSMWLV': 'qflag_DPOSMWLW_p_v',
    'DPOSSWLW': 'qflag_DPOSSWLW_p',
    'DPOSSWLV': 'qflag_DPOSSWLW_p_v',
    'DPOSSWMW': 'qflag_DPOSSWMW_p',
    'DPOSSWMV': 'qflag_DPOSSWMW_p_v',
    'HTHERPLW': 'qflag_HTHERPLW_p',
    'HTHERPLV': 'qflag_HTHERPLW_p_v',
    'HTHERPMW': 'qflag_HTHERPMW_p',
    'HTHERPMV': 'qflag_HTHERPMW_p_v',
    'HTHERPSW': 'qflag_HTHERPSW_p',
    'HTHERPSV': 'qflag_HTHERPSW_p_v',
    'HTHERSLW': 'qflag_HTHERSLW_p',
    'HTHERSLV': 'qflag_HTHERSLW_p_v',
    'HTHERSSW': 'qflag_HTHERSSW_p',
    'HTHERSSV': 'qflag_HTHERSSW_p_v',
    'INVTIMES': 'qflag_INVTIMES_p',
    'INVTIMEV': 'qflag_INVTIMES_p_v',
    'JOUTPLW': 'qflag_JOUTPLW_p',
    'JOUTPLV': 'qflag_JOUTPLW_p_v',
    'JOUTPMW': 'qflag_JOUTPMW_p',
    'JOUTPMV': 'qflag_JOUTPMW_p_v',
    'JOUTPSW': 'qflag_JOUTPSW_p',
    'JOUTPSV': 'qflag_JOUTPSW_p_v',
    'NOUTPLW': 'qflag_NOUTPLW_p',
    'NOUTPLV': 'qflag_NOUTPLW_p_v',
    'NOUTPMW': 'qflag_NOUTPMW_p',
    'NOUTPMV': 'qflag_NOUTPMW_p_v',
    'NOUTPSW': 'qflag_NOUTPSW_p',
    'NOUTPSV': 'qflag_NOUTPSW_p_v',
    'PLWGL1L': 'qflag_PLWGL1L_p',
    'PLWGL1V': 'qflag_PLWGL1L_p_v',
    'PMWGL1L': 'qflag_PMWGL1L_p',
    'PMWGL1V': 'qflag_PMWGL1L_p_v',
    'PSWGL1L': 'qflag_PSWGL1L_p',
    'PSWGL1V': 'qflag_PSWGL1L_p_v',
    'PLWGL2L': 'qflag_PLWGL2L_p',
    'PLWGL2V': 'qflag_PLWGL2L_p_v',
    'PMWGL2L': 'qflag_PMWGL2L_p',
    'PMWGL2V': 'qflag_PMWGL2L_p_v',
    'PSWGL2L': 'qflag_PSWGL2L_p',
    'PSWGL2V': 'qflag_PSWGL2L_p_v',
    'PLWOOCAL': 'qflag_PLWOOCAL_p',
    'PLWOOCAV': 'qflag_PLWOOCAL_p_v',
    'PMWOOCAL': 'qflag_PMWOOCAL_p',
    'PMWOOCAV': 'qflag_PMWOOCAL_p_v',
    'PSWOOCAL': 'qflag_PSWOOCAL_p',
    'PSWOOCAV': 'qflag_PSWOOCAL_p_v',
    'SLWOOCAL': 'qflag_SLWOOCAL_p',
    'SLWOOCAV': 'qflag_SLWOOCAL_p_v',
    'SSWOOCAL': 'qflag_SSWOOCAL_p',
    'SSWOOCAV': 'qflag_SSWOOCAL_p_v',
    'SLWRAT1L': 'qflag_SLWRAT1L_p',
    'SLWRAT1V': 'qflag_SLWRAT1L_p_v',
    'SSWRAT1L': 'qflag_SSWRAT1L_p',
    'SSWRAT1V': 'qflag_SSWRAT1L_p_v',
    'SLWRAT2L': 'qflag_SLWRAT2L_p',
    'SLWRAT2V': 'qflag_SLWRAT2L_p_v',
    'SSWRAT2L': 'qflag_SSWRAT2L_p',
    'SSWRAT2V': 'qflag_SSWRAT2L_p_v',
    'SLWIDENT': 'qflag_SLWIDENT_p',
    'SLWIDENV': 'qflag_SLWIDENT_p_v',
    'SSWIDENT': 'qflag_SSWIDENT_p',
    'SSWIDENV': 'qflag_SSWIDENT_p_v',
    'SLWDIFFR': 'qflag_SLWDIFFR_p',
    'SLWDIFFV': 'qflag_SLWDIFFR_p_v',
    'SSWDIFFR': 'qflag_SSWDIFFR_p',
    'SSWDIFFV': 'qflag_SSWDIFFR_p_v',
    'SLWOUTLR': 'qflag_SLWOUTLR_p',
    'SLWOUTLV': 'qflag_SLWOUTLR_p_v',
    'SSWOUTLR': 'qflag_SSWOUTLR_p',
    'SSWOUTLV': 'qflag_SSWOUTLR_p_v',
    'RATTRUNC': 'qflag_RATTRUNC_p',
    'RATTRUNV': 'qflag_RATTRUNC_p_v',
    'RATTRPLW': 'qflag_RATTRPLW_p',
    'RATTRPLV': 'qflag_RATTRPLW_p_v',
    'RATTRPMW': 'qflag_RATTRPMW_p',
    'RATTRPMV': 'qflag_RATTRPMW_p_v',
    'RATTRPSW': 'qflag_RATTRPSW_p',
    'RATTRPSV': 'qflag_RATTRPSW_p_v',
    'RATTRSLW': 'qflag_RATTRSLW_p',
    'RATTRSLV': 'qflag_RATTRSLW_p_v',
    'RATTRSSW': 'qflag_RATTRSSW_p',
    'RATTRSSV': 'qflag_RATTRSSW_p_v',
    'PHASEWRP': 'qflag_PHASEWRP_p',
    'PHASEWRV': 'qflag_PHASEWRP_p_v',
    'RETSPLIN': 'qflag_RETSPLIN_p',
    'SCANEXTR': 'qflag_SCANEXTR_p',
    'SCANEXTV': 'qflag_SCANEXTR_p_v',
    'SMECTEMP': 'qflag_SMECTEMP_p',
    'SMECTEMV': 'qflag_SMECTEMP_p_v',
    'AVGNGPSU': 'qflag_AVGNGPSU_p',
    'AVGNGPSV': 'qflag_AVGNGPSU_p_v',
    'AVNODPSU': 'qflag_AVNODPSU_p',
    'AVNODPSV': 'qflag_AVNODPSU_p_v',
    'NODPSU': 'qflag_NODPSU_p',
    'NODPSV': 'qflag_NODPSU_p_v',
    'AVGSPEED': 'qflag_AVGSPEED_p',
    'AVGSPEEV': 'qflag_AVGSPEED_p_v',
    'STDSPEED': 'qflag_STDSPEED_p',
    'STDSPEEV': 'qflag_STDSPEED_p_v',
    'WRBSESLW': 'qflag_WRBSESLW_p',
    'WRBSESLV': 'qflag_WRBSESLW_p_v',
    'WRBSESSW': 'qflag_WRBSESSW_p',
    'WRBSESSV': 'qflag_WRBSESSW_p_v',
    'BPSPUBUF': 'qflag_BPSPUBUF_p',
    'BPSPUBUV': 'qflag_BPSPUBUF_p_v',
    'BPSPUMIS': 'qflag_BPSPUMIS_p',
    'BPSPUMIV': 'qflag_BPSPUMIS_p_v',
    'BPSPUFAI': 'qflag_BPSPUFAI_p',
    'BPSPUFAV': 'qflag_BPSPUFAI_p_v',
    'RPSPUBUF': 'qflag_RPSPUBUF_p',
    'RPSPUBUV': 'qflag_RPSPUBUF_p_v',
    'RPSPUMIS': 'qflag_RPSPUMIS_p',
    'RPSPUMIV': 'qflag_RPSPUMIS_p_v',
    'RPSPUFAI': 'qflag_RPSPUFAI_p',
    'RPSPUFAV': 'qflag_RPSPUFAI_p_v',
    'BPSCISAT': 'qflag_BPSCISAT_p',
    'BPSCISAV': 'qflag_BPSCISAT_p_v',
    'RPSCISAT': 'qflag_RPSCISAT_p',
    'RPSCISAV': 'qflag_RPSCISAT_p_v',
    'BPCALSAT': 'qflag_BPCALSAT_p',
    'BPCALSAV': 'qflag_BPCALSAT_p_v',
    'RPCALSAT': 'qflag_RPCALSAT_p',
    'RPCALSAV': 'qflag_RPCALSAT_p_v',
    'BPSCGLIT': 'qflag_BPSCGLIT_p',
    'BPSCGLIV': 'qflag_BPSCGLIT_p_v',
    'RPSCGLIT': 'qflag_RPSCGLIT_p',
    'RPSCGLIV': 'qflag_RPSCGLIT_p_v',
    'BPCAGLIT': 'qflag_BPCAGLIT_p',
    'BPCAGLIV': 'qflag_BPCAGLIT_p_v',
    'RPCAGLIT': 'qflag_RPCAGLIT_p',
    'RPCAGLIV': 'qflag_RPCAGLIT_p_v',
    'POINTACC': 'qflag_POINTACC_p',
    'POINTACV': 'qflag_POINTACC_p_v',
    'POINTSTA': 'qflag_POINTSTA_p',
    'POINTSTV': 'qflag_POINTSTA_p_v',
    'POINTOFF': 'qflag_POINTOFF_p',
    'POINTOFV': 'qflag_POINTOFF_p_v',
    'BPSCNUMF': 'qflag_BPSCNUMF_p',
    'BPSCNUMV': 'qflag_BPSCNUMF_p_v',
    'BPSCRMSF': 'qflag_BPSCRMSF_p',
    'BPSCRMSV': 'qflag_BPSCRMSF_p_v',
    'BPSCVALF': 'qflag_BPSCVALF_p',
    'BPSCVALV': 'qflag_BPSCVALF_p_v',
    'BPSCVRMF': 'qflag_BPSCVRMF_p',
    'BPSCVRMV': 'qflag_BPSCVRMF_p_v',
    'RPSCNUMF': 'qflag_RPSCNUMF_p',
    'RPSCNUMV': 'qflag_RPSCNUMF_p_v',
    'RPSCRMSF': 'qflag_RPSCRMSF_p',
    'RPSCRMSV': 'qflag_RPSCRMSF_p_v',
    'RPSCVALF': 'qflag_RPSCVALF_p',
    'RPSCVALV': 'qflag_RPSCVALF_p_v',
    'RPSCVRMF': 'qflag_RPSCVRMF_p',
    'RPSCVRMV': 'qflag_RPSCVRMF_p_v',
    'BPCANUMF': 'qflag_BPCANUMF_p',
    'BPCANUMV': 'qflag_BPCANUMF_p_v',
    'BPCARMSF': 'qflag_BPCARMSF_p',
    'BPCARMSV': 'qflag_BPCARMSF_p_v',
    'BPCAVALF': 'qflag_BPCAVALF_p',
    'BPCAVALV': 'qflag_BPCAVALF_p_v',
    'BPCAVRMF': 'qflag_BPCAVRMF_p',
    'BPCAVRMV': 'qflag_BPCAVRMF_p_v',
    'RPCANUMF': 'qflag_RPCANUMF_p',
    'RPCANUMV': 'qflag_RPCANUMF_p_v',
    'RPCARMSF': 'qflag_RPCARMSF_p',
    'RPCARMSV': 'qflag_RPCARMSF_p_v',
    'RPCAVALF': 'qflag_RPCAVALF_p',
    'RPCAVALV': 'qflag_RPCAVALF_p_v',
    'RPCAVRMF': 'qflag_RPCAVRMF_p',
    'RPCAVRMV': 'qflag_RPCAVRMF_p_v',
    'BPVALPIX': 'qflag_BPVALPIX_p',
    'BPVALPIV': 'qflag_BPVALPIX_p_v',
    'RPVALPIX': 'qflag_RPVALPIX_p',
    'RPVALPIV': 'qflag_RPVALPIX_p_v',
    'BPOBSTRU': 'qflag_BPOBSTRU_p',
    'BPOBSTRV': 'qflag_BPOBSTRU_p_v',
    'RPOBSTRU': 'qflag_RPOBSTRU_p',
    'RPOBSTRV': 'qflag_RPOBSTRU_p_v',
    'BPSUBBCK': 'qflag_BPSUBBCK_p',
    'BPSUBBCV': 'qflag_BPSUBBCK_p_v',
    'RPSUBBCK': 'qflag_RPSUBBCK_p',
    'RPSUBBCV': 'qflag_RPSUBBCK_p_v',
    'FSPUBLUE': 'qflag_FSPUBLUE_p',
    'FSPUBLUV': 'qflag_FSPUBLUE_p_v',
    'FSPURED': 'qflag_FSPURED_p',
    'FSPUREV': 'qflag_FSPURED_p_v',
    'NOQDC': 'qflag_NOQDC_p',
    'FASTQDC': 'qflag_FASTQDC_p',
    'NOPOWCOR': 'qflag_NOPOWCOR_p',
    'FPUMIXC': 'qflag_FPUMIXC_p',
    'FPUMIXCV': 'qflag_FPUMIXCV_p',
    'FPUMIXV': 'qflag_FPUMIXV_p',
    'FPUMIXMC': 'qflag_FPUMIXMC_p',
    'FPUMIXMR': 'qflag_FPUMIXMR_p',
    'FPUCHOP': 'qflag_FPUCHOP_p',
    'FPUDPLX': 'qflag_FPUDPLX_p',
    'FPULNA': 'qflag_FPULNA_p',
    'FPUTMPH': 'qflag_FPUTMPH_p',
    'FPUTMPC': 'qflag_FPUTMPC_p',
    'FPUTMPL0': 'qflag_FPUTMPL0_p',
    'SPUR': 'qflag_SPUR_p',
    'POINTFAL': 'qflag_POINTFAL_p',
    'CHCKCOMB': 'qflag_CHCKCOMB_p',
    'CHCKZERO': 'qflag_CHCKZERO_p',
    'SPIKENUM': 'qflag_SPIKENUM_p',
    'SPIKENUV': 'qflag_SPIKENUM_p_v',
    'BADPIXEL': 'qflag_BADPIXEL_p',
    'BADPIXEV': 'qflag_BADPIXEL_p_v',
    'SATPIXEL': 'qflag_SATPIXEL_p',
    'SATPIXEV': 'qflag_SATPIXEL_p_v',
    'OBSMODE': 'qflag_OBSMODE_i',
    'MAXDRIFT': 'qflag_MAXDRIFT_p',
    'FREQCHEC': 'qflag_FREQCHEC_p',
    'CHOPATT': 'qflag_CHOPATT_p',
    'CHOPVAL': 'qflag_CHOPVAL_p',
    'LOFPATT': 'qflag_LOFPATT_p',
    'LOFVAL': 'qflag_LOFVAL_p',
    'LOFVAV': 'qflag_LOFVAL_p_v',
    'BUFFPATT': 'qflag_BUFFPATT_p',
    'BUFFVAL': 'qflag_BUFFVAL_p',
    'BUFFVAV': 'qflag_BUFFVAL_p_v',
    'PHASCHEC': 'qflag_PHASCHEC_p',
    'HOTCOLD': 'qflag_HOTCOLD_p',
    'TSYSFLAG': 'qflag_TSYSFLAG_p',
    'INTENCAL': 'qflag_INTENCAL_p',
    'NWEIGHTS': 'qflag_NWEIGHTS_p',
    'NOSUBST': 'qflag_NOSUBST_p',
    'NOBASELN': 'qflag_NOBASELN_p',
    'ONOFFSEQ': 'qflag_ONOFFSEQ_p',
    'ONOFFLEN': 'qflag_ONOFFLEN_p',
    'ONOFFPRO': 'qflag_ONOFFPRO_p',
    'NOFFSUBS': 'qflag_NOFFSUBS_p',
    'UNALIGHK': 'qflag_UNALIGHK_p',
    'UNALIGHV': 'qflag_UNALIGHK_p_v',
    'NOCHOPHK': 'qflag_NOCHOPHK_p',
    'NOCHOPHV': 'qflag_NOCHOPHK_p_v',
    'NOCOMMHK': 'qflag_NOCOMMHK_p',
    'NOCOMMHV': 'qflag_NOCOMMHK_p_v',
    'NOFREQHK': 'qflag_NOFREQHK_p',
    'NOFREQHV': 'qflag_NOFREQHK_p_v',
    'NOLOCOFF': 'qflag_NOLOCOFF_p',
    'NOLOCOFV': 'qflag_NOLOCOFF_p_v',
    'NOLOCOMA': 'qflag_NOLOCOMA_p',
    'NOLOCOMV': 'qflag_NOLOCOMA_p_v',
    'BBIDCORR': 'qflag_BBIDCORR_p',
    'BBIDCORV': 'qflag_BBIDCORR_p_v',
    'DFORDER': 'qflag_DFORDER_p',
    'LESSDATA': 'qflag_LESSDATA_p',
    'MOREDATA': 'qflag_MOREDATA_p',
    'RTERROR': 'qflag_RTERROR_p',
    'RTERROV': 'qflag_RTERROR_p_v',
    'CMDFAIL': 'qflag_CMDFAIL_p',
    'CMDFAIV': 'qflag_CMDFAIL_p_v',
    'TMLOSSES': 'qflag_TMLOSSES_p',
    'MIB': 'qflag_MIB_p',
    'EVENT': 'qflag_EVENT_p',
    'EXCPTION': 'qflag_EXCPTION_p',
    'ALARM': 'qflag_ALARM_p',
    'TCERRORS': 'qflag_TCERRORS_p',
    'SPOINTIN': 'qflag_SPOINTIN_p',
    'POINTING': 'qflag_POINTING_p',
    'DEGLITB': 'qflag_BSSCGLIT_p',
    'DEGLITBV': 'qflag_BSSCGLIT_p_v',
    'DEGLITR': 'qflag_RSSCGLIT_p',
    'DEGLITRV': 'qflag_RSSCGLIT_p_v',
    'ANOMY70': 'qflag_DMANOG1_p',
    'ANOMY70V': 'qflag_DMANOG1_p_v',
    'LOWPNTS': 'qflag_SLOWSAMP_p',
    'LOWPNTSV': 'qflag_SLOWSAMP_p_v',
    'NUMPIXB': 'qflag_BSVALPIX_p',
    'NUMPIXBV': 'qflag_BSVALPIX_p_v',
    'NUMPIXRB': 'qflag_RSVALPIX_p_v',
    'NUMPIXR': 'qflag_RSVALPIX_p',
    'SATUR_B': 'qflag_BSSCISAT_p',
    'SATUR_BV': 'qflag_BSSCISAT_p_v',
    'SATUR_R': 'qflag_RSSCISAT_p',
    'SATUR_RV': 'qflag_RSSCISAT_p_v',
    # HCSS-20946',
    'ANOMLPTG': 'qflag_ANOMLPTG_p',
    'BADDATDC': 'qflag_BADDATDC_p',
    'BADDATL2': 'qflag_BADDATL2_p',
    'BADHCDC': 'qflag_BADHCDC_p',
    'CALERR': 'qflag_CALERR_p',
    'CHKSINZR': 'qflag_CHKSINZR_p',
    'DARKSATR': 'qflag_DARKSATR_p',
    'DECONF': 'qflag_DECONF_p',
    'FPUDPLXC': 'qflag_FPUDPLXC_p',
    'GYRATTSP': 'qflag_GYRATTSP_p',
    'HD247194': 'qflag_HD247194_p',
    'HEBHAPPL': 'qflag_HEBHAPPL_p',
    'HEBVAPPL': 'qflag_HEBVAPPL_p',
    'HECRZERO': 'qflag_HECRZERO_p',
    'HM025193': 'qflag_HM025193_i',
    'HM029191': 'qflag_HM029191_i',
    'HM120191': 'qflag_HM120191_i',
    'HRSHASIC': 'qflag_HRSHASIC_p',
    'HRSVASIC': 'qflag_HRSVASIC_p',
    'HSDFAIL': 'qflag_HSDFAIL_p',
    'LOTUNING': 'qflag_LOTUNING_p',
    'LOUCUR': 'qflag_LOUCUR_p',
    'LOUDSB': 'qflag_LOUDSB_p',
    'LOUDSB1B': 'qflag_LOUDSB1B_p',
    'LRGTHROW': 'qflag_LRGTHROW_p',
    'NOFRQCAL': 'qflag_NOFRQCAL_p',
    'NOVELOC': 'qflag_NOVELOC_p',
    'PLTFRMNG': 'qflag_PLTFRMNG_p',
    'RMSHV': 'qflag_RMSHV_p',
    'RMSHVVST': 'qflag_RMSHVVST_p',
    'RMSNOISE': 'qflag_RMSNOISE_p',
    'SCANCNT': 'qflag_SCANCNT_p',
    'UNKBB': 'qflag_UNKBB_p',
    'WM409565': 'qflag_WM409565_i',
    'WM508565': 'qflag_WM508565_i',
    'WM608565': 'qflag_WM608565_i',
    'WRGFRQSC': 'qflag_WRGFRQSC_i',
    'WRONGAVG': 'qflag_WRONGAVG_p',
    'ZEROINCF': 'qflag_ZEROINCF_p',
    # HCSS-20580, 2015-09-01',
    'NRASTER': 'numPoints',
    'NLINES': 'numLines',
    # HCSS-21199',
    'PS3_0': 'ps3_0',
    'PS3_1': 'ps3_1',
    'PS3_2': 'ps3_2',
    # HCSS-21442',
    'ETAAH': 'apertureEfficiency_H',
    'ETAAV': 'apertureEfficiency_V',
    'ETAMBH': 'beamEff_H',
    'ETAMBV': 'beamEff_V',
    'UPCONVH': 'upConvert_H',
    'UPCONVV': 'upConvert_V',
    'SKREFCON': 'skyRefContam',
    'SKREFCOR': 'skyRefContamCorr',
    'ICREATOR': 'iccCreator',
    # HCSS-21442',
    'REFZEROV': 'zeroVelocityReference',
    'VELALGO': 'computeVelocityAlgorithm',
    'AREAPIX': 'areaPixels',
    'AREAAS': 'areaArcsecs',
    'ROWCENT': 'regionParameter_centerRow',
    'COLCENT': 'regionParameter_centerCol',
    'ROWRAD': 'regionParameter_radiusRow',
    'COLRAD': 'regionParameter_radiuscol',
    'ROWSTART': 'regionParameter_row1',
    'ROWEND': 'regionParameter_row2',
    'COLSTART': 'regionParameter_col1',
    'COLEND': 'regionParameter_col2',
    'CDESC3': 'cdesc3',
    'FLIPYX': 'flipxy',
    'ATTQUATR': 'attitudeQuaternion',
    'CHOP': 'Chopper',
    'BB_TYPE': 'bbtype',
    'NCHANNEL': 'channels',
    'CMDCHOP': 'cmd_chopper',
    'PRIMERED': 'prime_redundant',
    'LOFMEAS': 'LoFrequency_measured',
    'FRQWIDTH': 'frequencyWidth',
    'FRQMONIT': 'frequency_monitor',
    'LOFTHROW': 'loThrow',
    'CHKCOMB': 'checkComb',
    'COMBRES': 'resolution',
    'RES': 'resolution',
    'GAINMETH': 'gainMethod',
    'LOADINT': 'loadInterval',
    'BEAMUSED': 'beamUsed',
    'PIXELSIZ': 'pixelSize',
    'MAPSIZE': 'mapSize',
    'REFPIX': 'refPixelCoordinates',
    'PIXOFFS': 'pixelOffset',
    'FLYANGLE': 'flyAngle',
    'HPBWUSED': 'hpbwAssumed',
    'MAPWIDGR': 'mapWidthGridded',
    'MAPHEIGR': 'mapHeightGridded',
    'MAPWIDOB': 'mapWidthObserved',
    'MAPHEIOB': 'mapHeightObserved',
    'HEBCORRV': 'hebCorrVApplied',
    'HEBCORRH': 'hebCorrHApplied',
    'NDATASET': 'count_ds',
    'ERPFLAG': 'erpFlagged',
    # HCSS-21520',
    'GAINFLAG': 'gain_flag',
    'SPURREJC': 'spur_rejection',
    'STARTCOL': 'startCol',
    'ENDCOL': 'endCol',
    'STARTROW': 'startRow',
    'ENDROW': 'endRow',
    'ASPCTRAT': 'aspectRatio',
    'SLITWID': 'slitWidth',
    'ROWMIN': 'rowMin',
    'ROWMAX': 'rowMax',
    'COLMIN': 'colMin',
    'COLMAX': 'colMax',
}

# Dictionary of commonly used HCSS keywords that follow a numbering pattern
# e.g. OBSID123': 'obsid123, PROP1': 'proposal1
FITS_keywords_Numbered = {
    'OBSID': 'obsid',
    'PROP': 'proposal',
    'CUNIT': 'cunit',
    'CONSNAM': 'constraintName',
    'CONSTYP': 'constraintType',
    'CONSINF': 'constraintInfo',
    'CONS_ST': 'timeConstraintStart',
    'CONS_EN': 'timeConstraintEnd',
    'YARRAY': 'yArray',
    'ZARRAY': 'zArray',
    'PROCOI': 'propCoI',
    'PERTNAM': 'perturberName',
    'PERTGM': 'perturberGM',
    'CALTBC': 'calTableComment',
    # -- HCSS-19602 Deconvolution task',
    'LOF': 'loFreq',
    'NBD': 'num_bad_chan_in_scan',
    'BSC': 'bad_scan',
    'POBS_': 'photObsid',
}

FITS_KEYWORDS = copy.copy(FITS_keywords_Standard)
FITS_KEYWORDS.update(FITS_keywords_HEASARC)



# -*- coding: utf-8 -*-

import requests
import logging
import sys
from requests.auth import HTTPBasicAuth

from fdi.dataset.serializable import serialize
from fdi.dataset.deserialize import deserialize
from fdi.pal.urn import parseUrn, parse_poolurl
from .pnsconfig import pnsconfig as pcc
from fdi.utils.getconfig import getConfig

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    # strset = (str, unicode)
    strset = str
    from urlparse import urlparse


logger = logging.getLogger(__name__)
#logger.debug('level %d' % (logger.getEffectiveLevel()))


common_header = {
    'Accept': 'application/json',
    'Accept-Charset': 'utf-8',
    'Accept-Encoding': 'identity',
    'Cache-Control': 'no-cache',
    'Connection': 'keep-alive',
    "Content-type": 'application/json'
}


pcc.update(getConfig())
defaulturl = 'http://' + pcc['node']['host'] + \
    ':' + str(pcc['node']['port']) + pcc['baseurl']


def urn2fdiurl(urn, poolurl, contents='product', method='GET'):
    """ Returns URL for accessing pools with a URN.

    This is done by using the PoolURL.

    contents:
    'product' for returning a product from the pool.
    'hk' for returning the housekeeping data of the pool.
    'classes' for returning the class housekeeping data of the pool.
    'urns' for returning the URN housekeeping data of the pool.
    'tags' for returning the tag housekeeping data of the pool.

    method:
    'GET' compo for retrieving product or hk or classes, urns, tags,
    'POST' compo for uploading  product
    DELETE compo for removing product or pool

    Example:
    IP=ip poolpath=/a poolname=b/c files=/a/b/c/classes.jsn | urns.jsn | t.. | urn...

    with python:
    m.refs['myinput'] = special_ref
    ref=pstore.save(m)
    assert ref.urn == 'urn:c:fdi.dataset.MapContext:203'
    p=ref.product
    myref=p.refs['myinput']

    with a pool:
    myref=pool.load('http://ip:port/v0.6/a/b/c/fdi.dataset.MapContext/203/refs/myinput')

    At the same time this is not allowed due to overlapping after 'c'

    urn:http://ip:port/a/b/d/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/a/b/c/d/

    This is also overlapping starting from '/'

    urn:http://ip:port/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/

    but for example this is allowed:
    urn:http://ip:port/a/k/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/a/k/
    """

    poolname, resourcecn, index = parseUrn(urn)
    indexs = str(index)
    poolpath, scheme, place, pn = parse_poolurl(poolurl, poolhint=poolname)
    if poolname is None:
        poolname = pn
    if method == 'GET':
        if contents == 'product':
            ret = poolurl + '/' + resourcecn + '/' + indexs
        elif contents == 'housekeeping':
            ret = poolurl + '/hk'
        elif contents in ['classes', 'urns', 'tags']:
            ret = poolurl + '/hk/' + contents
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    elif method == 'POST':
        if contents == 'product':
            ret = poolurl + '/' + resourcecn + '/' + indexs
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    elif method == 'DELETE':
        if contents == 'pool':
            ret = poolurl
        elif contents == 'product':
            ret = poolurl + '/' + resourcecn + '/' + indexs
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    else:
        raise ValueError(method)
    return ret

# Store tag in headers, maybe that's  not a good idea


def save_to_server(data, urn, poolurl, tag):
    """Save product to server with putting tag in headers
    """
    user = pcc['auth_user']
    password = pcc['auth_pass']
    auth = HTTPBasicAuth(user, password)
    api = urn2fdiurl(urn, poolurl, contents='product', method='POST')
    # print('POST API: ' + api)
    headers = {'tag': tag}
    res = requests.post(
        api, auth=auth, data=serialize(data), headers=headers)
    result = deserialize(res.text)
    # print(result)
    return result


def read_from_server(urn, poolurl, contents='product'):
    """Read product or hk data from server
    """
    user = pcc['auth_user']
    password = pcc['auth_pass']
    auth = HTTPBasicAuth(user, password)
    api = urn2fdiurl(urn, poolurl, contents=contents)
    # print("GET REQUEST API: " + api)
    res = requests.get(api, auth=auth)
    result = deserialize(res.text)
    return result['result'], result['msg']


def delete_from_server(urn, poolurl, contents='product'):
    """Remove a product or pool from server
    """
    user = pcc['auth_user']
    password = pcc['auth_pass']
    auth = HTTPBasicAuth(user, password)
    api = urn2fdiurl(urn, poolurl, contents=contents, method='DELETE')
    # print("DELETE REQUEST API: " + api)
    res = requests.delete(api, auth=auth)
    result = deserialize(res.text)
    return result['result'], result['msg']



#!flask/bin/python
# -*- coding: utf-8 -*-

from fdi.pns.pnsconfig import pnsconfig as pc
from fdi.utils.options import opt
from fdi.utils.getconfig import getConfig

#sys.path.insert(0, abspath(join(join(dirname(__file__), '..'), '..')))

# print(sys.path)


def setuplogging():
    import logging.config
    import logging
    from fdi.pns import logdict  # import logdict
    # create logger
    logging.config.dictConfig(logdict.logdict)
    return logging


logging = setuplogging()
logger = logging.getLogger(__name__)


if __name__ == '__main__':

    logger = logging.getLogger()
    # default configuration is provided. Copy pnsconfig.py to ~/.config/pnslocal.py
    pc.update(getConfig())
    logger.setLevel(pc['logginglevel'])
    logger.info('Server starting. Make sure no other instance is running')

    node = pc['node']
    # Get username and password and host ip and port.
    ops = [
        {'long': 'help', 'char': 'h', 'default': False, 'description': 'print help'},
        {'long': 'verbose', 'char': 'v', 'default': False,
            'description': 'print info'},
        {'long': 'username=', 'char': 'u',
            'default': node['username'], 'description':'user name/ID'},
        {'long': 'password=', 'char': 'p',
            'default': node['password'], 'description':'password'},
        {'long': 'host=', 'char': 'i',
            'default': node['host'], 'description':'host IP/name'},
        {'long': 'port=', 'char': 'o',
            'default': node['port'], 'description':'port number'},
        {'long': 'server=', 'char': 's',
            'default': 'pns', 'description': 'server type: pns or httppool_server'},

    ]

    out = opt(ops)
    verbose = out[1]['result']
    for j in range(2, 6):
        n = out[j]['long'].strip('=')
        node[n] = out[j]['result']
    servertype = out[6]['result']

    if verbose:
        logger.setLevel(logging.DEBUG)

    logger.info('logging level %d' % (logger.getEffectiveLevel()))
    if node['username'] in ['', None] or node['password'] in ['', None]:
        logger.error(
            'Error. Specify non-empty username and password on commandline')
        exit(3)
    print('Check http://' + node['host'] + ':' + str(node['port']) +
          pc['baseurl'] + '/ for API list')

    if servertype == 'pns':
        print('======== %s ========' % servertype)
        from fdi.pns.pns_server import app
    elif servertype == 'httppool_server':
        print('<<<<<< %s >>>>>' % servertype)
        from fdi.pns.httppool_server import app
    else:
        logger.error('Unknown server %s' % servertype)
        sys.exit(-1)

    app.run(host=node['host'], port=node['port'],
            threaded=True, debug=verbose, processes=1, use_reloader=False)



# -*- coding: utf-8 -*-

from ..utils.masked import masked
from ..utils.common import grouper
from .serializable import Serializable
from .datatypes import DataTypes, DataTypeNames
from .odict import ODict
from .composite import Composite
from .listener import DatasetEventSender, ParameterListener, DatasetListener, DatasetEvent, EventTypeOf
from .quantifiable import Quantifiable
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable
from .finetime import FineTime, FineTime1, utcobj
from .classes import Classes
from .typed import Typed
from .invalid import INVALID
from ..utils.common import exprstrs, wls, mstr, t2l

from tabulate import tabulate

import builtins
from collections import OrderedDict
from numbers import Number
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


ParameterClasses = {
    'AbstractParameter': dict(value=None,
                              description='UNKNOWN'),

    'Parameter': dict(value=None,
                      description='UNKNOWN',
                      typ_='',
                      default=None,
                      valid=None),

    'NumericParameter': dict(value=None,
                             description='UNKNOWN',
                             typ_='',
                             unit=None,
                             default=None,
                             valid=None,
                             typecode=None),

    'DateParameter': dict(value=None,
                          description='UNKNOWN',
                          default='',
                          valid=None,
                          typecode=None),

    'StringParameter': dict(value=None,
                            description='UNKNOWN',
                            default='',
                            valid=None,
                            typecode='B'),

}


def parameterDataClasses(tt):
    """ maps machine type names to class objects """
    if tt not in DataTypeNames:
        raise TypeError("Type %s is not in %s." %
                        (tt, str([''.join(x) for x in DataTypeNames])))
    if tt == 'int':
        return int
    elif tt in builtins.__dict__:
        return builtins.__dict__[tt]
    else:
        return Classes.mapping[tt]


class AbstractParameter(Annotatable, Copyable, DeepEqual, DatasetEventSender, Serializable):
    """ Parameter is the interface for all named attributes
    in the MetaData container. It can have a value and a description.
    Default     value=None, description='UNKNOWN'
    """

    def __init__(self, value=None, description='UNKNOWN', **kwds):
        """ Constructed with no argument results in a parameter of
        None value and 'UNKNOWN' description ''.
        With a signle argument: arg -> value, 'UNKNOWN' as default-> description.
        With two positional arguments: arg1-> value, arg2-> description.
        Type is set according to value's.
        Unsuported parameter types will get a NotImplementedError.
        """
        super(AbstractParameter, self).__init__(
            description=description, **kwds)

        self.setValue(value)

    def accept(self, visitor):
        """ Adds functionality to classes of this type."""
        visitor.visit(self)

    @property
    def value(self):
        """ for property getter
        """
        return self.getValue()

    @value.setter
    def value(self, value):
        """ for property setter
        """
        self.setValue(value)

    def getValue(self):
        """ Gets the value of this parameter as an Object. """
        return self._value

    def setValue(self, value):
        """ Replaces the current value of this parameter.
        """
        self._value = value

    def __setattr__(self, name, value):
        """ add eventhandling """
        super(AbstractParameter, self).__setattr__(name, value)

        # this will fail during init when annotatable init sets description
        # if issubclass(self.__class__, DatasetEventSender):
        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, \
                EventType.UNKNOWN_ATTRIBUTE_CHANGED, \
                (name, value), None, None

            nu = name.upper()
            if nu in EventTypeOf['CHANGED']:
                ty = EventTypeOf['CHANGED'][nu]
            else:
                tv = EventType.UNKNOWN_ATTRIBUTE_CHANGED
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)

#    def ff(self, name, value):
#
#        if eventType is not None:
#            if eventType not in EventType:
#                # return eventType
#                raise ValueError(str(eventType))
#            elif eventType != EventType.UNKOWN_ATTRIBUTE_CHANGED:
#                # super() has found the type
#                return eventType
#        # eventType is None or is UNKOWN_ATTRIBUTE_CHANGED
#            if name == 'value':
#                ty = EventType.VALUE_CHANGED
#                ch = (value)
#            elif name == 'description':
#                ty = EventType.DESCRIPTION_CHANGED
#            else:
#                # raise AttributeError(
#                #    'Parameter "'+self.description + '" has no attribute named '+name)
#                pass
#            if ty != EventType.UNKOWN_ATTRIBUTE_CHANGED:
#                e = DatasetEvent(source=so, target=ta, typ_=ty,
#                                 change=ch, cause=ca, rootCause=ro)
#                self.fire(e)
#            return ty
#        return eventType
#
    def __eq__(self, obj, verbose=False, **kwds):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return self.value == obj
        else:
            return super(AbstractParameter, self).__eq__(obj)

    def __lt__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return self.value < obj
        else:
            return super(AbstractParameter, self).__lt__(obj)

    def __gt__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return self.value > obj
        else:
            return super(AbstractParameter, self).__gt__(obj)

    def __le__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return self.value <= obj
        else:
            return super(AbstractParameter, self).__le__(obj)

    def __ge__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return self.value >= obj
        else:
            return super(AbstractParameter, self).__ge__(obj)

    def getValueAsString():
        """ Value as string for building the string representation of the parameter.  """
        return

    def __repr__(self):
        return self.toString(level=1)

    def toString(self, level=0, alist=False, **kwds):
        """ alist: returns a dictionary string representation of parameter attributes.
        """
        vs = str(self._value)
        ds = str(self.description)
        ss = '%s' % (vs) if level else \
            '%s, "%s"' % (vs, ds)
        if alist:
            return exprstrs(self)
        return self.__class__.__name__ + ss

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           value=self.value,
                           listeners=self.listeners,
                           _STID=self._STID
                           )

    __hash__ = DeepEqual.__hash__


class Parameter(AbstractParameter, Typed):
    """ Parameter is the interface for all named attributes
    in the MetaData container. It can have a value and a description.
    Defaul arguments: typ_='', default=None, valid=None.
    value=default, description='UNKNOWN'
    """

    def __init__(self, value=None, description='UNKNOWN', typ_='', default=None, valid=None, **kwds):
        """ invoked with no argument results in a parameter of
        None value and 'UNKNOWN' description ''. typ_ DataTypes[''], which is None.
        With a signle argument: arg -> value, 'UNKNOWN'-> description. ParameterTypes-> typ_, hex values have integer typ_.
        Unsuported parameter types will get a NotImplementedError.
f        With two positional arguments: arg1-> value, arg2-> description. ParameterTypes['']-> typ_.
        Unsuported parameter types will get a NotImplementedError.
        With three positional arguments: arg1 casted to DataTypes[arg3]-> value, arg2-> description. arg3-> typ_.
        Unsuported parameter types will get a NotImplementedError.
        Incompatible value and typ_ will get a TypeError.
        """
        # super(Parameter, self).__init__(description=description, **kwds)

        self.setDefault(default)
        self.setValid(valid)
        # super() will set value so type and default need to be set first
        super(Parameter, self).__init__(
            value=value, description=description, typ_=typ_, **kwds)

    def accept(self, visitor):
        """ Adds functionality to classes of this type."""
        visitor.visit(self)

    def setType(self, typ_):
        """ Replaces the current type of this parameter.
        Defaul will be casted if not the same.
        Unsuported parameter types will get a NotImplementedError.
        """
        if typ_ is None or typ_ == '':
            self._type = ''
            return
        if typ_ in DataTypes:
            super().setType(typ_)
            # let setdefault deal with type
            self.setDefault(self._default)
        else:
            raise NotImplementedError(
                'Parameter type %s is not in %s.' %
                (typ_, str([''.join(x) for x in DataTypes])))

    def checked(self, value):
        """ Checks input value against self.type.

        If value is none, returns it;
        else if type is not set, return value after setting type;
        If value's type is a subclass of self's type, return the value;
        If value's and self's types are both subclass of Number, returns value casted in self's type.
        """
        if not hasattr(self, '_type'):

            return value

        t_type = type(value)
        t = t_type.__name__
        st = self._type
        if st == '' or st is None:
            # self does not have a type
            try:
                ct = DataTypeNames[t]
                if ct == 'vector':
                    self._type = 'quaternion' if len(value) == 4 else ct
                else:
                    self._type = ct
            except KeyError as e:
                raise TypeError("Type %s is not in %s." %
                                (t, str([''.join(x) for x in DataTypeNames])))
            return value

        # self has type
        tt = DataTypes[st]
        if tt in Classes.mapping:
            # custom-defined parameter. delegate checking to themselves
            if issubclass(type(value), tuple):
                # frozendict used in baseproduct module change lists to tuples
                # which causes deserialized parameter to differ from ProductInfo.
                value = list(value)
            return value
        tt_type = builtins.__dict__[tt]
        if issubclass(t_type, tt_type):
            return value
        elif issubclass(t_type, Number) and issubclass(tt_type, Number):
            # , if both are Numbers.Number, value is casted into given typ_.
            return tt_type(value)
            # st = tt
        else:
            vs = hex(value) if t == 'int' and st == 'hex' else str(value)
            raise TypeError(
                'Value %s is of type %s, but should be %s.' % (vs, t, tt))

    @property
    def default(self):
        return self.getDefault()

    @default.setter
    def default(self, default):
        self.setDefault(default)

    def getDefault(self):
        """ Returns the default related to this object."""
        return self._default

    def setDefault(self, default):
        """ Sets the default of this object.

        Default is set directly if type is not set or default is None.
        If the type of default is not getType(), TypeError is raised.
        """

        if default is None:
            self._default = default
            return

        self._default = self.checked(default)

    @property
    def valid(self):
        return self.getValid()

    @valid.setter
    def valid(self, valid):
        self.setValid(valid)

    def getValid(self):
        """ Returns the valid related to this object."""
        return self._valid

    def setValid(self, valid):
        """ Sets the valid of this object.

        If valid is None or empty, set as None, else save in a way so the tuple keys can be serialized with JSON.
        """

        self._valid = None if valid is None or len(
            valid) == 0 else [t2l([k, v]) for k, v in valid.items()] if issubclass(valid.__class__, dict) else t2l(valid)

    def isValid(self):
        res = self.validate(self.value)
        if issubclass(res.__class__, tuple):
            return res[0] is not INVALID
        else:
            return True

    def split(self, into=None):
        """ split a multiple binary bit-masked parameters according to masks.

        into: dictionary mapping bit-masks to the sub-name of the parameter.
        return: a dictionary mapping name of new parameters to its value.
        """
        ruleset = self.getValid()
        if ruleset is None or len(ruleset) == 0:
            return {}

        st = DataTypes[self._type]
        vt = type(self._value).__name__

        if st is not None and st != '' and vt != st:
            return {}

        masks = {}
        # number of bits of mask
        highest = 0
        for rn in ruleset:
            rule, name = tuple(rn)
            if issubclass(rule.__class__, (tuple, list)):
                if rule[0] is Ellipsis or rule[1] is Ellipsis:
                    continue
                if rule[0] >= rule[1]:
                    # binary masked rules are [mask, vld] e.g. [0B011000,0b11]
                    mask, valid_val = rule[0], rule[1]
                    masked_val, mask_height, mask_width = masked(
                        self._value, mask)
                    masks[mask] = masked_val
                    if mask_height > highest:
                        highest = mask_height

        if into is None or len(into) < len(masks):
            # like {'0b110000': 0b10, '0b001111': 0b0110}
            fmt = '#0%db' % (highest + 2)
            return {format(mask, fmt): value for mask, value in masks.items()}
        else:
            # use ``into`` for rulename
            # like {'foo': 0b10, 'bar': 0b0110}
            return {into[mask]: value for mask, value in masks.items()}

    def validate(self, value=INVALID):
        """ checks if a match the rule set.

        value: will be checked against the ruleset. Default is ``self._valid``.
        returns:
        (valid value, rule name) for discrete and range rules.
        {mask: (valid val, rule name, mask_height, mask_width), ...} for binary masks rules.
        (INVALID, 'Invalid') if no matching is found.
        (value, 'Default') if rule set is empty.
        """

        if value is INVALID:
            value = self._value

        ruleset = self.getValid()
        if ruleset is None or len(ruleset) == 0:
            return (value, 'Default')

        st = DataTypes[self._type]
        vt = type(value).__name__

        if st is not None and st != '' and vt != st:
            return (INVALID, 'Type '+vt)

        binmasks = {}
        hasvalid = False
        for rn in ruleset:
            rule, name = tuple(rn)
            res = INVALID
            if issubclass(rule.__class__, (tuple, list)):
                if rule[0] is Ellipsis:
                    res = INVALID if (value > rule[1]) else value
                elif rule[1] is Ellipsis:
                    res = INVALID if (value < rule[0]) else value
                elif rule[0] >= rule[1]:
                    # binary masked rules are [mask, vld] e.g. [0B011000,0b11]
                    mask, vld = rule[0], rule[1]
                    if len(binmasks.setdefault(mask, [])) == 0:
                        vtest, mask_height, mask_width = masked(value, mask)
                        if vtest == vld:
                            # record, indexed by mask
                            binmasks[mask] += [vld, name,
                                               mask_height, mask_width]
                else:
                    # range
                    res = INVALID if (value < rule[0]) or (
                        value > rule[1]) else value
            else:
                # discrete value
                res = value if rule == value else INVALID
            if not hasvalid:
                # record the 1st valid
                if res is not INVALID:
                    hasvalid = (res, name)
        if any(len(resnm) for mask, resnm in binmasks.items()):
            return [tuple(resnm) if len(resnm) else (INVALID, 'Invalid') for mask, resnm in binmasks.items()]
        return hasvalid if hasvalid else (INVALID, 'Invalid')

    def setValue(self, value):
        """ Replaces the current value of this parameter.

        If value is None set it to default.
        If given/current type is '' and arg value's type is in DataTypes both value and type are updated to the suitable one in DataTypeNames; or else TypeError is raised.
        If value type and given/current type are different.
            Incompatible value and type will get a TypeError.
        """

        if value is None:
            self._value = self._default if hasattr(self, '_default') else value
            return
        self._value = self.checked(value)

    def __repr__(self):
        return self.toString(level=1)

    def toString(self, level=0, alist=False, **kwds):

        ret = exprstrs(self, level=level, **kwds)
        if alist:
            return ret
        vs, us, ts, ds, fs, gs, cs = ret
        if level:
            return vs
        return '%s{ %s <%s>, "%s", default= %s, valid= %s tcode=%s}' % \
            (self.__class__.__name__, vs, ts, ds, fs, gs, cs)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           value=self._value,
                           type=self._type,
                           default=self._default,
                           valid=self._valid,
                           listeners=self.listeners,
                           _STID=self._STID
                           )

    __hash__ = DeepEqual.__hash__


class NumericParameter(Parameter, Quantifiable):
    """ has a number as the value, a unit, and a typecode.
    """

    def __init__(self, value=None, description='UNKNOWN', typ_='', default=None, valid=None, **kwds):
        super(NumericParameter, self).__init__(
            value=value, description=description, typ_=typ_, default=default, valid=valid, **kwds)

    def __repr__(self):
        return self.toString(level=1)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           value=self._value,
                           type=self._type,
                           default=self._default,
                           valid=self._valid,
                           unit=self._unit,
                           typecode=self._typecode,
                           _STID=self._STID)

    __hash__ = DeepEqual.__hash__


class DateParameter(Parameter):
    """ has a FineTime as the value.
    """

    def __init__(self, value=None, description='UNKNOWN', default=0, valid=None, typecode=None, **kwds):
        """
        if value and typecode are both given, typecode will be overwritten by value.format.
        """
        self.setTypecode(typecode)
        # this will set default then set value.
        super(DateParameter, self).__init__(
            value=value, description=description, typ_='finetime', default=default, valid=valid, **kwds)

    @ property
    def typecode(self):
        return self.getTypecode()

    @ typecode.setter
    def typecode(self, typecode):
        self.setTypecode(typecode)

    def getTypecode(self):
        """ Returns the typecode related to this object. None if value not set."""
        if not hasattr(self, '_value'):
            return None
        return self._value.format

    def setTypecode(self, typecode):
        """ Sets the typecode of this object. quietly returns if value not set."""
        if not hasattr(self, '_value'):
            return
        self._value.format = typecode

    def setValue(self, value):
        """ accept any type that a FineTime does.
        """
        if value is not None and not issubclass(value.__class__, FineTime):
            value = FineTime(date=value, format=self.getTypecode())
        super().setValue(value)

    def setDefault(self, default):
        """ accept any type that a FineTime does.
        """
        if default is not None and not issubclass(default.__class__, FineTime):
            default = FineTime(date=default, format=self.getTypecode())
        super().setDefault(default)

    def __repr__(self):

        return self.toString(level=1)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           value=self._value,
                           default=self._default,
                           valid=self._valid,
                           typecode=self.typecode,
                           _STID=self._STID)
        return self.__class__.__name__ + \
            '{ description = "%s", value = "%s", typecode = "%s"}' % \
            (str(self.description), str(self.value), str(self.getTypecode()))

    __hash__ = DeepEqual.__hash__


class DateParameter1(DateParameter):
    """ Like DateParameter but usese  FineTime. """

    def setValue(self, value):
        """ accept any type that a FineTime1 does.
        """
        if value is not None and not issubclass(value.__class__, FineTime1):
            value = FineTime1(date=value, format=self.getTypecode())
        super().setValue(value)

    def setDefault(self, default):
        """ accept any type that a FineTime1 does.
        """
        if default is not None and not issubclass(default.__class__, FineTime1):
            default = FineTime1(date=default, format=self.getTypecode())
        super().setDefault(default)

    __hash__ = DeepEqual.__hash__


class StringParameter(Parameter):
    """ has a unicode string as the value, a typecode for length and char.
    """

    def __init__(self, value=None, description='UNKNOWN', valid=None, default='', typecode='B', **kwds):
        self.setTypecode(typecode)
        super(StringParameter, self).__init__(
            value=value, description=description, typ_='string', default=default, valid=valid, **kwds)

    @ property
    def typecode(self):
        return self.getTypecode()

    @ typecode.setter
    def typecode(self, typecode):
        self.setTypecode(typecode)

    def getTypecode(self):
        """ Returns the typecode related to this object."""
        return self._typecode

    def setTypecode(self, typecode):
        """ Sets the typecode of this object. """
        self._typecode = typecode

    def __repr__(self):
        return self.toString(level=1)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(value=self._value,
                           description=self.description,
                           valid=self._valid,
                           default=self._default,
                           typecode=self._typecode,
                           _STID=self._STID)

    __hash__ = DeepEqual.__hash__


MetaHeaders = ['name', 'value', 'unit', 'type', 'valid',
               'default', 'code', 'description']


class MetaData(Composite, Copyable, Serializable, ParameterListener, DatasetEventSender):
    """ A container of named Parameters. A MetaData object can
    have one or more parameters, each of them stored against a
    unique name. The order of adding parameters to this container
    is important, that is: the keySet() method will return a set of
    labels of the parameters in the sequence as they were added.
    Note that replacing a parameter with the same name,
    will keep the order. """

    def __init__(self, copy=None, **kwds):
        super(MetaData, self).__init__(**kwds)
        if copy:
            # not implemented ref https://stackoverflow.com/questions/10640642/is-there-a-decent-way-of-creating-a-copy-constructor-in-python
            logger.error('use copy.copy() insteadof MetaData(copy)')
        else:
            return

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern."""
        visitor.visit(self)

    def clear(self):
        """ Removes all the key - parameter mappings. """
        self.getDataWrappers().clear()

    def set(self, name, newParameter):
        """ Saves the parameter and  add eventhandling.
        Raises TypeError if not given Parameter (sub) class object.
        """
        if not issubclass(newParameter.__class__, AbstractParameter):
            raise TypeError('Only Parameters can be saved.')

        super(MetaData, self).set(name, newParameter)

        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, -1, \
                (name, newParameter), None, None
            if name in self.keySet():
                ty = EventType.PARAMETER_CHANGED
            else:
                ty = EventType.PARAMETER_ADDED
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)

    def remove(self, name):
        """ add eventhandling """
        r = super(MetaData, self).remove(name)
        if r is None:
            return r

        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, -1, \
                (name), None, None  # generic initial vals
            ty = EventType.PARAMETER_REMOVED
            ch = (name, r)
            # raise ValueError('Attempt to remove non-existant parameter "%s"' % (name))
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)
        return r

    def toString(self, level=0, tablefmt='grid', tablefmt1='simple',
                 widths=None, **kwds):
        """ return  string representation of metada.

        level: 0 is the most detailed, 2 is the least, 1 is for __repr__.
        tablefmt: format string in packae ``tabulate``.
        widths: controls how the attributes of every parameter are displayed in the table cells. If is set to -1, there is no cell-width limit. For finer control set a dictionary of parameter attitute names and how many characters wide its tsble cell is, 0 for ommiting the attributable. Default is
``{'name': 8, 'value': 17, 'unit': 7, 'type': 8,
                      'valid': 25, 'default': 17, 'code': 4, 'description': 15}``
        """

        if widths is None:
            widths = {'name': 8, 'value': 17, 'unit': 7, 'type': 8,
                      'valid': 25, 'default': 17, 'code': 4, 'description': 15}
        tab = []
        # N parameters per row for level 1
        N = 3
        i, row = 0, []
        cn = self.__class__.__name__
        s = ''
        att = {}
        for (k, v) in self._sets.items():
            att['name'] = str(k)
            att['value'], att['unit'], att['type'], att['description'],\
                att['default'], att['valid'], att['code'] = v.toString(
                    alist=True)
            if level == 0:
                l = tuple(att[n] for n in MetaHeaders) if widths == -1 else \
                    tuple(wls(att[n], w) for n, w in widths.items() if w != 0)
                tab.append(l)
            elif level == 1:
                ps = '%s= %s' % (att['name'], att['value'])
                # s += mstr(self, level=level, depth=1,  tablefmt = tablefmt, **kwds)
                tab.append(wls(ps, 80//N))
                if 0:
                    row.append(wls(ps, 80//N))
                    i += 1
                    if i == N:
                        tab.append(row)
                        i, row = 0, []
            else:
                tab.append(att['name'])

        lsnr = self.listeners.toString(level=level, **kwds)

        if level == 0:
            headers = MetaHeaders if widths == -1 else \
                [n for n, w in zip(MetaHeaders, widths) if w != 0]
            fmt = tablefmt
            s += tabulate(tab, headers=headers, tablefmt=fmt, missingval='',
                          disable_numparse=True)
        elif level == 1:
            t = grouper(tab, N)
            headers = ''
            fmt = tablefmt1
            s += tabulate(t, headers=headers, tablefmt=fmt, missingval='',
                          disable_numparse=True)
        else:
            s += ', '.join(tab)
            return '%s, listeners = %s' % (s, lsnr)
        return '\n%s\n%s-listeners = %s' % (s, cn, lsnr) if len(tab) else \
            '%s%s-listeners = %s' % ('(No parameter.)', cn, lsnr)

    def __repr__(self, **kwds):
        return self.toString(level=1, **kwds)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        # print(self.listeners)
        # print([id(o) for o in self.listeners])

        return OrderedDict(_sets=self._sets,
                           listeners=self.listeners,
                           _STID=self._STID)
    __hash__ = DeepEqual.__hash__
